from typing import TypeAlias

from litellm.types.llms.openai import (
    ChatCompletionAnnotation,
    ChatCompletionAnnotationURLCitation,
)

from ..typing.content import Content
from ..typing.message import (
    AssistantMessage,
    SystemMessage,
    ToolMessage,
    UserMessage,
)
from ..typing.tool import ToolCall
from . import (
    OpenAIAssistantMessageParam,
    OpenAICompletionMessage,
    OpenAIDeveloperMessageParam,
    OpenAIFunctionMessageParam,
    OpenAISystemMessageParam,
    OpenAIToolCallFunction,
    OpenAIToolCallParam,
    OpenAIToolMessageParam,
    OpenAIUserMessageParam,
)
from .content_converters import from_api_content, to_api_content

OpenAIMessageType: TypeAlias = (
    OpenAIAssistantMessageParam
    | OpenAIToolMessageParam
    | OpenAIUserMessageParam
    | OpenAIDeveloperMessageParam
    | OpenAISystemMessageParam
    | OpenAIFunctionMessageParam
)


def from_api_user_message(
    api_message: OpenAIUserMessageParam, name: str | None = None
) -> UserMessage:
    content = from_api_content(api_message["content"])
    name = api_message.get("name")

    return UserMessage(content=content, name=name)


def to_api_user_message(message: UserMessage) -> OpenAIUserMessageParam:
    api_content = (
        to_api_content(message.content)
        if isinstance(message.content, Content)
        else message.content
    )
    api_name = message.name
    api_message = OpenAIUserMessageParam(role="user", content=api_content)
    if api_name is not None:
        api_message["name"] = api_name

    return api_message


def from_api_assistant_message(
    api_message: OpenAICompletionMessage, name: str | None = None
) -> AssistantMessage:
    tool_calls = None
    if api_message.tool_calls is not None:
        tool_calls = [
            ToolCall(
                id=tool_call.id,
                tool_name=tool_call.function.name,  # type: ignore
                tool_arguments=tool_call.function.arguments,  # type: ignore
            )
            for tool_call in api_message.tool_calls
        ]

    annotations = None
    if api_message.annotations is not None:
        annotations = [
            ChatCompletionAnnotation(
                type="url_citation",
                url_citation=ChatCompletionAnnotationURLCitation(
                    **api_annotation.url_citation.model_dump()
                ),
            )
            for api_annotation in api_message.annotations
        ]

    return AssistantMessage(
        content=api_message.content,
        tool_calls=tool_calls,
        refusal=api_message.refusal,
        annotations=annotations,
        name=name,
    )


def to_api_assistant_message(
    message: AssistantMessage,
) -> OpenAIAssistantMessageParam:
    api_tool_calls = None
    if message.tool_calls is not None:
        api_tool_calls = [
            OpenAIToolCallParam(
                type="function",
                id=tool_call.id,
                function=OpenAIToolCallFunction(
                    name=tool_call.tool_name,
                    arguments=tool_call.tool_arguments,
                ),
            )
            for tool_call in message.tool_calls
        ]

    api_message = OpenAIAssistantMessageParam(role="assistant", content=message.content)

    if message.name is not None:
        api_message["name"] = message.name
    if api_tool_calls is not None:
        api_message["tool_calls"] = api_tool_calls or []
    if message.refusal is not None:
        api_message["refusal"] = message.refusal

    # TODO: avoid this hack
    if message.content is None:
        # Some API providers return None in the generated content without errors,
        # even though None in the input content is not accepted.
        api_message["content"] = "<empty>"

    return api_message


def from_api_system_message(
    api_message: OpenAISystemMessageParam, name: str | None = None
) -> SystemMessage:
    return SystemMessage(content=api_message["content"], name=name)  # type: ignore


def to_api_system_message(message: SystemMessage) -> OpenAISystemMessageParam:
    api_message = OpenAISystemMessageParam(role="system", content=message.content)
    if message.name is not None:
        api_message["name"] = message.name

    return api_message


def from_api_tool_message(
    api_message: OpenAIToolMessageParam, name: str | None = None
) -> ToolMessage:
    return ToolMessage(
        content=api_message["content"],  # type: ignore
        tool_call_id=api_message["tool_call_id"],
        name=name,
    )


def to_api_tool_message(message: ToolMessage) -> OpenAIToolMessageParam:
    return OpenAIToolMessageParam(
        role="tool", content=message.content, tool_call_id=message.tool_call_id
    )



from collections.abc import AsyncIterator, Sequence
from itertools import pairwise
from logging import getLogger
from typing import Any, Generic, Protocol, TypeVar, cast, final

from grasp_agents.tracing_decorators import workflow

from ..errors import WorkflowConstructionError
from ..packet_pool import Packet
from ..processors.base_processor import BaseProcessor
from ..run_context import CtxT, RunContext
from ..typing.events import Event, ProcPacketOutputEvent, WorkflowResultEvent
from ..typing.io import InT, OutT, ProcName
from .workflow_processor import WorkflowProcessor

logger = getLogger(__name__)

_OutT_contra = TypeVar("_OutT_contra", contravariant=True)


class WorkflowLoopTerminator(Protocol[_OutT_contra, CtxT]):
    def __call__(
        self,
        out_packet: Packet[_OutT_contra],
        ctx: RunContext[CtxT] | None,
        **kwargs: Any,
    ) -> bool: ...


class LoopedWorkflow(WorkflowProcessor[InT, OutT, CtxT], Generic[InT, OutT, CtxT]):
    def __init__(
        self,
        name: ProcName,
        subprocs: Sequence[BaseProcessor[Any, Any, Any, CtxT]],
        exit_proc: BaseProcessor[Any, OutT, Any, CtxT],
        recipients: list[ProcName] | None = None,
        max_retries: int = 0,
        max_iterations: int = 10,
    ) -> None:
        super().__init__(
            subprocs=subprocs,
            name=name,
            start_proc=subprocs[0],
            end_proc=exit_proc,
            recipients=recipients,
            max_retries=max_retries,
        )

        for prev_proc, proc in pairwise(subprocs):
            if prev_proc.out_type != proc.in_type:
                raise WorkflowConstructionError(
                    f"Output type {prev_proc.out_type} of subprocessor "
                    f"{prev_proc.name} does not match input type {proc.in_type} of "
                    f"subprocessor {proc.name}"
                )
        if subprocs[-1].out_type != subprocs[0].in_type:
            raise WorkflowConstructionError(
                "Looped workflow's last subprocessor output type "
                f"{subprocs[-1].out_type} does not match first subprocessor input "
                f"type {subprocs[0].in_type}"
            )

        self._max_iterations = max_iterations

        self.workflow_loop_terminator: WorkflowLoopTerminator[OutT, CtxT] | None
        if not hasattr(type(self), "workflow_loop_terminator"):
            self.workflow_loop_terminator = None

    @property
    def max_iterations(self) -> int:
        return self._max_iterations

    def add_workflow_loop_terminator(
        self, func: WorkflowLoopTerminator[OutT, CtxT]
    ) -> WorkflowLoopTerminator[OutT, CtxT]:
        self.workflow_loop_terminator = func

        return func

    def _terminate_workflow_loop(
        self,
        out_packet: Packet[OutT],
        *,
        ctx: RunContext[CtxT] | None = None,
        **kwargs: Any,
    ) -> bool:
        if self.workflow_loop_terminator:
            return self.workflow_loop_terminator(out_packet, ctx=ctx, **kwargs)

        return False

    @workflow(name="workflow_run")  # type: ignore
    @final
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | Sequence[InT] | None = None,
        call_id: str | None = None,
        forgetful: bool = False,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT]:
        packet = in_packet
        exit_packet: Packet[OutT] | None = None
        call_id = self._generate_call_id(call_id)

        for iteration_num in range(1, self._max_iterations + 1):
            for subproc in self.subprocs:
                logger.info(f"\n[Running subprocessor {subproc.name}]\n")

                packet = await subproc.run(
                    chat_inputs=chat_inputs,
                    in_packet=packet,
                    in_args=in_args,
                    forgetful=forgetful,
                    call_id=f"{call_id}/{subproc.name}/iter_{iteration_num}",
                    ctx=ctx,
                )

                logger.info(f"\n[Finished running subprocessor {subproc.name}]\n")

                if subproc is self._end_proc:
                    exit_packet = cast("Packet[OutT]", packet)
                    if self._terminate_workflow_loop(exit_packet, ctx=ctx):
                        return exit_packet
                    if iteration_num == self._max_iterations:
                        logger.info(
                            f"Max iterations reached ({self._max_iterations}). "
                            "Exiting loop."
                        )
                        return exit_packet

                chat_inputs = None
                in_args = None

        raise RuntimeError("Looped workflow did not exit after max iterations.")

    @workflow(name="workflow_run")  # type: ignore
    @final
    async def run_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | Sequence[InT] | None = None,
        call_id: str | None = None,
        forgetful: bool = False,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncIterator[Event[Any]]:
        packet = in_packet
        exit_packet: Packet[OutT] | None = None
        call_id = self._generate_call_id(call_id)

        for iteration_num in range(1, self._max_iterations + 1):
            for subproc in self.subprocs:
                logger.info(f"\n[Running subprocessor {subproc.name}]\n")

                async for event in subproc.run_stream(
                    chat_inputs=chat_inputs,
                    in_packet=packet,
                    in_args=in_args,
                    forgetful=forgetful,
                    call_id=f"{call_id}/{subproc.name}/iter_{iteration_num}",
                    ctx=ctx,
                ):
                    if isinstance(event, ProcPacketOutputEvent):
                        packet = event.data
                    yield event

                logger.info(f"\n[Finished running subprocessor {subproc.name}]\n")

                if subproc is self._end_proc:
                    exit_packet = cast("Packet[OutT]", packet)
                    if self._terminate_workflow_loop(exit_packet, ctx=ctx):
                        yield WorkflowResultEvent(
                            data=exit_packet, proc_name=self.name, call_id=call_id
                        )
                        return
                    if iteration_num == self._max_iterations:
                        logger.info(
                            f"Max iterations reached ({self._max_iterations}). "
                            "Exiting loop."
                        )
                        yield WorkflowResultEvent(
                            data=exit_packet, proc_name=self.name, call_id=call_id
                        )
                        return

                chat_inputs = None
                in_args = None



from collections.abc import Sequence
from typing import Any, Protocol

from pydantic import PrivateAttr

from .memory import Memory
from .run_context import RunContext
from .typing.io import LLMPrompt
from .typing.message import Message, Messages, SystemMessage


class MemoryPreparator(Protocol):
    def __call__(
        self,
        memory: "LLMAgentMemory",
        *,
        in_args: Any | None,
        sys_prompt: LLMPrompt | None,
        ctx: RunContext[Any],
        call_id: str,
    ) -> None: ...


class LLMAgentMemory(Memory):
    _message_history: Messages = PrivateAttr(default_factory=Messages)
    _sys_prompt: LLMPrompt | None = PrivateAttr(default=None)

    def __init__(self, sys_prompt: LLMPrompt | None = None) -> None:
        super().__init__()
        self._sys_prompt = sys_prompt
        self.reset(sys_prompt)

    @property
    def sys_prompt(self) -> LLMPrompt | None:
        return self._sys_prompt

    @property
    def message_history(self) -> Messages:
        return self._message_history

    def reset(
        self, sys_prompt: LLMPrompt | None = None, ctx: RunContext[Any] | None = None
    ):
        if sys_prompt is not None:
            self._sys_prompt = sys_prompt

        self._message_history = (
            [SystemMessage(content=self._sys_prompt)]
            if self._sys_prompt is not None
            else []
        )

    def erase(self) -> None:
        self._message_history = []

    def update(
        self, messages: Sequence[Message], *, ctx: RunContext[Any] | None = None
    ):
        self._message_history.extend(messages)

    @property
    def is_empty(self) -> bool:
        return len(self._message_history) == 0

    def __repr__(self) -> str:
        return (
            "LLMAgentMemory with message history of "
            f"length {len(self._message_history)}"
        )



import hashlib
import json
import logging
import sys
from collections.abc import AsyncIterator, Mapping, Sequence
from typing import Any, Literal, TypeAlias

from pydantic import BaseModel
from termcolor import colored
from termcolor._types import Color

from grasp_agents.typing.completion_chunk import CompletionChunk
from grasp_agents.typing.events import (
    AnnotationsChunkEvent,
    AnnotationsEndEvent,
    AnnotationsStartEvent,
    CompletionChunkEvent,
    # CompletionEndEvent,
    CompletionStartEvent,
    Event,
    GenMessageEvent,
    MessageEvent,
    ProcPacketOutputEvent,
    ResponseChunkEvent,
    ResponseEndEvent,
    ResponseStartEvent,
    RunResultEvent,
    SystemMessageEvent,
    ThinkingChunkEvent,
    ThinkingEndEvent,
    ThinkingStartEvent,
    ToolCallChunkEvent,
    ToolCallEndEvent,
    ToolCallStartEvent,
    ToolMessageEvent,
    UserMessageEvent,
    WorkflowResultEvent,
)

from .typing.completion import Usage
from .typing.content import Content, ContentPartText
from .typing.message import (
    AssistantMessage,
    Message,
    Role,
    SystemMessage,
    ToolMessage,
    UserMessage,
)

logger = logging.getLogger(__name__)


ROLE_TO_COLOR: Mapping[Role, Color] = {
    Role.SYSTEM: "magenta",
    Role.USER: "green",
    Role.ASSISTANT: "light_blue",
    Role.TOOL: "blue",
}

AVAILABLE_COLORS: list[Color] = [
    "magenta",
    "green",
    "light_blue",
    "light_cyan",
    "yellow",
    "blue",
    "red",
]

ColoringMode: TypeAlias = Literal["agent", "role"]
CompletionBlockType: TypeAlias = Literal["response", "thinking", "tool_call"]


def stream_colored_text(new_colored_text: str) -> None:
    sys.stdout.write(new_colored_text)
    sys.stdout.flush()


def get_color(
    agent_name: str = "", role: Role = Role.ASSISTANT, color_by: ColoringMode = "role"
) -> Color:
    if color_by == "agent":
        idx = int(
            hashlib.md5(agent_name.encode()).hexdigest(),  # noqa :S324
            16,
        ) % len(AVAILABLE_COLORS)

        return AVAILABLE_COLORS[idx]
    return ROLE_TO_COLOR[role]


def content_to_str(content: Content | str | None, role: Role) -> str:
    if role == Role.USER and isinstance(content, Content):
        content_str_parts: list[str] = []
        for content_part in content.parts:
            if isinstance(content_part, ContentPartText):
                content_str_parts.append(content_part.data.strip(" \n"))
            elif content_part.data.type == "url":
                content_str_parts.append(str(content_part.data.url))
            elif content_part.data.type == "base64":
                content_str_parts.append("<ENCODED_IMAGE>")
        return "\n".join(content_str_parts)

    assert isinstance(content, str | None)

    return (content or "").strip(" \n")


def truncate_content_str(content_str: str, trunc_len: int = 2000) -> str:
    if len(content_str) > trunc_len:
        return content_str[:trunc_len] + "[...]"

    return content_str


class Printer:
    def __init__(
        self,
        color_by: ColoringMode = "role",
        msg_trunc_len: int = 20000,
        output_to: Literal["print", "log"] = "print",
        logging_level: Literal["info", "debug", "warning", "error"] = "info",
    ) -> None:
        self.color_by: ColoringMode = color_by
        self.msg_trunc_len = msg_trunc_len
        self._current_message: str = ""
        self._logging_level = logging_level
        self._output_to = output_to

    def print_message(
        self,
        message: Message,
        agent_name: str,
        call_id: str,
        usage: Usage | None = None,
    ) -> None:
        if usage is not None and not isinstance(message, AssistantMessage):
            raise ValueError(
                "Usage information can only be printed for AssistantMessage"
            )
        color = get_color(
            agent_name=agent_name, role=message.role, color_by=self.color_by
        )
        log_kwargs = {"extra": {"color": color}}

        out = f"<{agent_name}> [{call_id}]\n"

        # Thinking
        if isinstance(message, AssistantMessage) and message.reasoning_content:
            thinking = message.reasoning_content.strip(" \n")
            out += f"<thinking>\n{thinking}\n</thinking>\n"

        # Content
        content = content_to_str(message.content or "", message.role)
        if content:
            try:
                content = json.dumps(json.loads(content), indent=2)
            except Exception:
                pass
            content = truncate_content_str(content, trunc_len=self.msg_trunc_len)
            if isinstance(message, SystemMessage):
                out += f"<system>\n{content}\n</system>\n"
            elif isinstance(message, UserMessage):
                out += f"<input>\n{content}\n</input>\n"
            elif isinstance(message, AssistantMessage):
                out += f"<response>\n{content}\n</response>\n"
            else:
                out += f"<tool result> [{message.tool_call_id}]\n{content}\n</tool result>\n"

        # Tool calls
        if isinstance(message, AssistantMessage) and message.tool_calls is not None:
            for tool_call in message.tool_calls:
                out += (
                    f"<tool call> {tool_call.tool_name} [{tool_call.id}]\n"
                    f"{tool_call.tool_arguments}\n</tool call>\n"
                )

        # Usage
        if usage is not None:
            usage_str = f"I/O/R/C tokens: {usage.input_tokens}/{usage.output_tokens}"
            usage_str += f"/{usage.reasoning_tokens or '-'}"
            usage_str += f"/{usage.cached_tokens or '-'}"

            out += f"\n------------------------------------\n{usage_str}\n"

        if self._output_to == "log":
            if self._logging_level == "debug":
                logger.debug(out, **log_kwargs)  # type: ignore
            elif self._logging_level == "info":
                logger.info(out, **log_kwargs)  # type: ignore
            elif self._logging_level == "warning":
                logger.warning(out, **log_kwargs)  # type: ignore
            else:
                logger.error(out, **log_kwargs)  # type: ignore
        else:
            stream_colored_text(colored(out + "\n", color))

    def print_messages(
        self,
        messages: Sequence[Message],
        agent_name: str,
        call_id: str,
        usages: Sequence[Usage | None] | None = None,
    ) -> None:
        _usages: Sequence[Usage | None] = usages or [None] * len(messages)

        for _message, _usage in zip(messages, _usages, strict=False):
            self.print_message(
                _message, usage=_usage, agent_name=agent_name, call_id=call_id
            )


async def print_event_stream(
    event_generator: AsyncIterator[Event[Any]],
    color_by: ColoringMode = "role",
    trunc_len: int = 10000,
) -> AsyncIterator[Event[Any]]:
    def _make_chunk_text(event: CompletionChunkEvent[CompletionChunk]) -> str:
        color = get_color(
            agent_name=event.proc_name or "", role=Role.ASSISTANT, color_by=color_by
        )
        text = ""

        if isinstance(event, CompletionStartEvent):
            text += f"\n<{event.proc_name}> [{event.call_id}]\n"
        elif isinstance(event, ThinkingStartEvent):
            text += "<thinking>\n"
        elif isinstance(event, ResponseStartEvent):
            text += "<response>\n"
        elif isinstance(event, ToolCallStartEvent):
            tc = event.data.tool_call
            text += f"<tool call> {tc.tool_name} [{tc.id}]\n"
        elif isinstance(event, AnnotationsStartEvent):
            text += "<annotations>\n"

        # if isinstance(event, CompletionEndEvent):
        #     text += f"\n</{event.proc_name}>\n"
        if isinstance(event, ThinkingEndEvent):
            text += "\n</thinking>\n"
        elif isinstance(event, ResponseEndEvent):
            text += "\n</response>\n"
        elif isinstance(event, ToolCallEndEvent):
            text += "\n</tool call>\n"
        elif isinstance(event, AnnotationsEndEvent):
            text += "\n</annotations>\n"

        if isinstance(event, ThinkingChunkEvent):
            thinking = event.data.thinking
            if isinstance(thinking, str):
                text += thinking
            else:
                text = "\n".join(
                    [block.get("thinking", "[redacted]") for block in thinking]
                )

        if isinstance(event, ResponseChunkEvent):
            text += event.data.response

        if isinstance(event, ToolCallChunkEvent):
            text += event.data.tool_call.tool_arguments or ""

        if isinstance(event, AnnotationsChunkEvent):
            text += "\n".join(
                [
                    json.dumps(annotation, indent=2)
                    for annotation in event.data.annotations
                ]
            )

        return colored(text, color)

    def _make_message_text(
        event: MessageEvent[SystemMessage | UserMessage | ToolMessage],
    ) -> str:
        message = event.data
        role = message.role
        content = content_to_str(message.content, role=role)

        color = get_color(
            agent_name=event.proc_name or "", role=role, color_by=color_by
        )
        text = f"\n<{event.proc_name}> [{event.call_id}]\n"

        if isinstance(event, (SystemMessageEvent, UserMessageEvent)):
            content = truncate_content_str(content, trunc_len=trunc_len)

        if isinstance(event, SystemMessageEvent):
            text += f"<system>\n{content}\n</system>\n"

        elif isinstance(event, UserMessageEvent):
            text += f"<input>\n{content}\n</input>\n"

        elif isinstance(event, ToolMessageEvent):
            message = event.data
            try:
                content = json.dumps(json.loads(content), indent=2)
            except Exception:
                pass
            text += (
                f"<tool result> [{message.tool_call_id}]\n{content}\n</tool result>\n"
            )

        return colored(text, color)

    def _make_packet_text(
        event: ProcPacketOutputEvent | WorkflowResultEvent | RunResultEvent,
    ) -> str:
        if isinstance(event, WorkflowResultEvent):
            src = "workflow"
        elif isinstance(event, RunResultEvent):
            src = "run"
        else:
            src = "processor"

        color = get_color(
            agent_name=event.proc_name or "", role=Role.ASSISTANT, color_by=color_by
        )
        text = f"\n<{event.proc_name}> [{event.call_id}]\n"

        if event.data.payloads:
            text += f"<{src} output>\n"
            for p in event.data.payloads:
                if isinstance(p, BaseModel):
                    p_str = p.model_dump_json(indent=2)
                else:
                    try:
                        p_str = json.dumps(p, indent=2)
                    except TypeError:
                        p_str = str(p)
                text += f"{p_str}\n"
            text += f"</{src} output>\n"

        return colored(text, color)

    # ------ Wrap event generator -------

    async for event in event_generator:
        yield event

        if isinstance(event, CompletionChunkEvent) and isinstance(
            event.data, CompletionChunk
        ):
            stream_colored_text(_make_chunk_text(event))

        if isinstance(event, MessageEvent) and not isinstance(event, GenMessageEvent):
            stream_colored_text(_make_message_text(event))

        if isinstance(
            event, (ProcPacketOutputEvent, WorkflowResultEvent, RunResultEvent)
        ):
            stream_colored_text(_make_packet_text(event))



import ast
import asyncio
import json
import re
from collections.abc import AsyncIterator, Coroutine, Mapping
from datetime import UTC, datetime
from logging import getLogger
from pathlib import Path
from typing import Annotated, Any, NoReturn, TypeVar, get_args, get_origin

from pydantic import TypeAdapter
from pydantic import ValidationError as PydanticValidationError
from tqdm.autonotebook import tqdm

from .errors import JSONSchemaValidationError, PyJSONStringParsingError

logger = getLogger(__name__)

_JSON_START_RE = re.compile(r"[{\[]")


def extract_json_substring(text: str) -> str | None:
    decoder = json.JSONDecoder()
    for match in _JSON_START_RE.finditer(text):
        start = match.start()
        try:
            _, end = decoder.raw_decode(text, idx=start)
            return text[start:end]
        except json.JSONDecodeError:
            continue

    return None


def parse_json_or_py_string(
    s: str,
    from_substring: bool = False,
    return_none_on_failure: bool = False,
    strip_language_markdown: bool = True,
) -> dict[str, Any] | list[Any] | None:
    s_orig = s

    if strip_language_markdown:
        s = re.sub(r"```[a-zA-Z0-9]*\n|```", "", s).strip()

    if from_substring:
        s = extract_json_substring(s) or ""

    try:
        return ast.literal_eval(s)
    except (ValueError, SyntaxError):
        try:
            return json.loads(s)
        except json.JSONDecodeError as exc:
            err_message = (
                "Both ast.literal_eval and json.loads "
                f"failed to parse the following JSON/Python string:\n{s_orig}"
            )
            if return_none_on_failure:
                logger.warning(err_message)
                return None
            raise PyJSONStringParsingError(s_orig, message=err_message) from exc


def is_str_type(t: Any) -> bool:
    type_origin = get_origin(t)
    type_args = get_args(t)

    return (t is str) or (
        (type_origin is Annotated) and len(type_args) > 0 and type_args[0] is str
    )


T = TypeVar("T")


def validate_obj_from_json_or_py_string(
    s: str,
    schema: type[T],
    from_substring: bool = False,
    strip_language_markdown: bool = True,
) -> T:
    try:
        if is_str_type(schema):
            parsed = s
        else:
            parsed = parse_json_or_py_string(
                s,
                return_none_on_failure=True,
                from_substring=from_substring,
                strip_language_markdown=strip_language_markdown,
            )
        return TypeAdapter(schema).validate_python(parsed)
    except PydanticValidationError as exc:
        raise JSONSchemaValidationError(s, schema) from exc


def validate_tagged_objs_from_json_or_py_string(
    s: str,
    schema_by_xml_tag: Mapping[str, type[T]],
    from_substring: bool = False,
    strip_language_markdown: bool = True,
) -> Mapping[str, T]:
    validated_obj_per_tag: dict[str, T] = {}
    _schema: type[T] | None = None
    _tag: str | None = None

    def _raise(
        tag: str,
        schema: Any,
        *,
        not_found: bool = False,
        base_exc: Exception | None = None,
    ) -> NoReturn:
        if not_found:
            msg = (
                f"Failed to find valid tagged section <{tag}>...</{tag}> "
                f"in the string:\n{s}"
            )
        else:
            msg = (
                f"Failed to validate substring within tag <{tag}> against JSON schema:"
                f"\n{s}\nExpected type: {schema}"
            )
        raise JSONSchemaValidationError(s, schema, message=msg) from base_exc

    for _tag, _schema in schema_by_xml_tag.items():
        if f"<{_tag}>" in s:
            match = re.search(rf"<{_tag}>\s*(.*?)\s*</{_tag}>", s, re.DOTALL)
            if match is None:
                _raise(_tag, _schema, not_found=True, base_exc=None)
            try:
                tagged_substring = match.group(1).strip()
                validated_obj_per_tag[_tag] = validate_obj_from_json_or_py_string(
                    tagged_substring,  # type: ignore[assignment]
                    schema=_schema,
                    from_substring=from_substring,
                    strip_language_markdown=strip_language_markdown,
                )
            except JSONSchemaValidationError as exc:
                _raise(_tag, _schema, base_exc=exc)

    return validated_obj_per_tag


def extract_xml_list(text: str) -> list[str]:
    pattern = re.compile(r"<(chunk_\d+)>(.*?)</\1>", re.DOTALL)

    chunks: list[str] = []
    for match in pattern.finditer(text):
        content = match.group(2).strip()
        chunks.append(content)
    return chunks


def read_txt(file_path: str | Path, encoding: str = "utf-8") -> str:
    return Path(file_path).read_text(encoding=encoding)


def read_contents_from_file(
    file_path: str | Path,
    binary_mode: bool = False,
) -> str | bytes:
    try:
        if binary_mode:
            return Path(file_path).read_bytes()
        return Path(file_path).read_text()
    except FileNotFoundError:
        logger.exception(f"File {file_path} not found.")
        return ""


def get_prompt(prompt_text: str | None, prompt_path: str | Path | None) -> str | None:
    if prompt_text is None:
        return read_contents_from_file(prompt_path) if prompt_path is not None else None  # type: ignore[arg-type]

    return prompt_text


async def asyncio_gather_with_pbar(
    *corouts: Coroutine[Any, Any, Any],
    no_tqdm: bool = False,
    desc: str | None = None,
) -> list[Any]:
    # TODO: optimize
    pbar = tqdm(total=len(corouts), desc=desc, disable=no_tqdm)

    async def run_and_update(coro: Coroutine[Any, Any, Any]) -> Any:
        result = await coro
        pbar.update(1)
        return result

    wrapped_tasks = [run_and_update(c) for c in corouts]
    results = await asyncio.gather(*wrapped_tasks)
    pbar.close()

    return results


def get_timestamp() -> str:
    return datetime.now(UTC).strftime("%Y%m%d_%H%M%S")


_T = TypeVar("_T")


async def stream_concurrent(
    generators: list[AsyncIterator[_T]],
) -> AsyncIterator[tuple[int, _T]]:
    queue: asyncio.Queue[tuple[int, _T] | None] = asyncio.Queue()
    pumps_left = len(generators)

    async def pump(gen: AsyncIterator[_T], idx: int) -> None:
        nonlocal pumps_left
        try:
            async for item in gen:
                await queue.put((idx, item))
        finally:
            pumps_left -= 1
            if pumps_left == 0:
                await queue.put(None)

    for idx, gen in enumerate(generators):
        asyncio.create_task(pump(gen, idx))

    while True:
        msg = await queue.get()
        if msg is None:
            break
        yield msg



import asyncio
import json
from collections.abc import AsyncIterator, Coroutine, Mapping, Sequence
from itertools import starmap
from logging import getLogger
from typing import Any, Generic, Protocol, final

from pydantic import BaseModel

from grasp_agents.tracing_decorators import task
from grasp_agents.typing.completion_chunk import CompletionChunk

from .errors import AgentFinalAnswerError
from .llm import LLM
from .llm_agent_memory import LLMAgentMemory
from .run_context import CtxT, RunContext
from .typing.completion import Completion
from .typing.events import (
    CompletionChunkEvent,
    CompletionEvent,
    Event,
    GenMessageEvent,
    LLMStreamingErrorEvent,
    ToolCallEvent,
    ToolMessageEvent,
    UserMessageEvent,
)
from .typing.message import AssistantMessage, Messages, ToolMessage, UserMessage
from .typing.tool import BaseTool, NamedToolChoice, ToolCall, ToolChoice

logger = getLogger(__name__)


class ToolCallLoopTerminator(Protocol[CtxT]):
    def __call__(
        self,
        conversation: Messages,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> bool: ...


class MemoryManager(Protocol[CtxT]):
    def __call__(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> None: ...


class LLMPolicyExecutor(Generic[CtxT]):
    def __init__(
        self,
        *,
        agent_name: str,
        llm: LLM,
        tools: list[BaseTool[BaseModel, Any, CtxT]] | None,
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        max_turns: int,
        react_mode: bool = False,
        final_answer_type: type[BaseModel] = BaseModel,
        final_answer_as_tool_call: bool = False,
    ) -> None:
        super().__init__()

        self._agent_name = agent_name

        self._final_answer_type = final_answer_type
        self._final_answer_as_tool_call = final_answer_as_tool_call
        self._final_answer_tool = self.get_final_answer_tool()

        tools_list: list[BaseTool[BaseModel, Any, CtxT]] | None = tools
        if tools and final_answer_as_tool_call:
            tools_list = tools + [self._final_answer_tool]
        self._tools = {t.name: t for t in tools_list} if tools_list else None

        self._response_schema = response_schema
        self._response_schema_by_xml_tag = response_schema_by_xml_tag

        self._llm = llm

        self._max_turns = max_turns
        self._react_mode = react_mode

        self.tool_call_loop_terminator: ToolCallLoopTerminator[CtxT] | None = None
        self.memory_manager: MemoryManager[CtxT] | None = None

    @property
    def agent_name(self) -> str:
        return self._agent_name

    @property
    def llm(self) -> LLM:
        return self._llm

    @property
    def response_schema(self) -> Any | None:
        return self._response_schema

    @response_schema.setter
    def response_schema(self, value: Any | None) -> None:
        self._response_schema = value

    @property
    def response_schema_by_xml_tag(self) -> Mapping[str, Any] | None:
        return self._response_schema_by_xml_tag

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, Any, CtxT]]:
        return self._tools or {}

    @property
    def max_turns(self) -> int:
        return self._max_turns

    @final
    def _terminate_tool_call_loop(
        self,
        conversation: Messages,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> bool:
        if self.tool_call_loop_terminator:
            return self.tool_call_loop_terminator(
                conversation, ctx=ctx, call_id=call_id, **kwargs
            )

        return False

    @final
    def _manage_memory(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> None:
        if self.memory_manager:
            self.memory_manager(memory=memory, ctx=ctx, call_id=call_id, **kwargs)

    @task(name="generate_message")  # type: ignore
    async def generate_message(
        self,
        memory: LLMAgentMemory,
        *,
        tool_choice: ToolChoice | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> AssistantMessage:
        completion = await self.llm.generate_completion(
            memory.message_history,
            response_schema=self.response_schema,
            response_schema_by_xml_tag=self.response_schema_by_xml_tag,
            tools=self.tools,
            tool_choice=tool_choice,
            proc_name=self.agent_name,
            call_id=call_id,
        )
        memory.update(completion.messages)
        self._process_completion(completion, ctx=ctx, call_id=call_id)

        return completion.messages[0]

    @task(name="generate_message")  # type: ignore
    async def generate_message_stream(
        self,
        memory: LLMAgentMemory,
        *,
        tool_choice: ToolChoice | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> AsyncIterator[
        CompletionChunkEvent[CompletionChunk]
        | CompletionEvent
        | GenMessageEvent
        | LLMStreamingErrorEvent
    ]:
        completion: Completion | None = None

        llm_event_stream = self.llm.generate_completion_stream(
            memory.message_history,
            response_schema=self.response_schema,
            response_schema_by_xml_tag=self.response_schema_by_xml_tag,
            tools=self.tools,
            tool_choice=tool_choice,
            proc_name=self.agent_name,
            call_id=call_id,
        )
        llm_event_stream_post = self.llm.postprocess_event_stream(llm_event_stream)  # type: ignore[assignment]

        async for event in llm_event_stream_post:
            if isinstance(event, CompletionEvent):
                completion = event.data
            yield event
        if completion is None:
            return

        yield GenMessageEvent(
            proc_name=self.agent_name, call_id=call_id, data=completion.messages[0]
        )

        memory.update(completion.messages)

        self._process_completion(completion, ctx=ctx, call_id=call_id)

    # @task(name="call_tools")  # type: ignore
    async def call_tools(
        self,
        calls: Sequence[ToolCall],
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> Sequence[ToolMessage]:
        # TODO: Add image support
        corouts: list[Coroutine[Any, Any, BaseModel]] = []
        for call in calls:
            tool = self.tools[call.tool_name]
            args = json.loads(call.tool_arguments)
            corouts.append(tool(ctx=ctx, call_id=call_id, **args))

        outs = await asyncio.gather(*corouts)
        tool_messages = list(
            starmap(ToolMessage.from_tool_output, zip(outs, calls, strict=True))
        )

        memory.update(tool_messages)

        if ctx.printer:
            ctx.printer.print_messages(
                tool_messages, agent_name=self.agent_name, call_id=call_id
            )

        return tool_messages

    # @task(name="call_tools")  # type: ignore
    async def call_tools_stream(
        self,
        calls: Sequence[ToolCall],
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> AsyncIterator[ToolMessageEvent]:
        tool_messages = await self.call_tools(
            calls, memory=memory, ctx=ctx, call_id=call_id
        )
        for tool_message, call in zip(tool_messages, calls, strict=True):
            yield ToolMessageEvent(
                proc_name=call.tool_name, call_id=call_id, data=tool_message
            )

    def _extract_final_answer_from_tool_calls(
        self, memory: LLMAgentMemory
    ) -> AssistantMessage | None:
        last_message = memory.message_history[-1]
        if not isinstance(last_message, AssistantMessage):
            return None

        for tool_call in last_message.tool_calls or []:
            if tool_call.tool_name == self._final_answer_tool.name:
                final_answer_message = AssistantMessage(
                    name=self.agent_name, content=tool_call.tool_arguments
                )
                last_message.tool_calls = None
                memory.update([final_answer_message])

                return final_answer_message

    @task(name="enforce_final_answer")  # type: ignore
    async def _generate_final_answer(
        self, memory: LLMAgentMemory, ctx: RunContext[CtxT], call_id: str
    ) -> AssistantMessage:
        user_message = UserMessage.from_text(
            "Exceeded the maximum number of turns: provide a final answer now!"
        )
        memory.update([user_message])
        if ctx.printer:
            ctx.printer.print_messages(
                [user_message], agent_name=self.agent_name, call_id=call_id
            )

        tool_choice = NamedToolChoice(name=self._final_answer_tool.name)
        await self.generate_message(
            memory, tool_choice=tool_choice, ctx=ctx, call_id=call_id
        )

        final_answer_message = self._extract_final_answer_from_tool_calls(memory=memory)
        if final_answer_message is None:
            raise AgentFinalAnswerError(proc_name=self.agent_name, call_id=call_id)

        return final_answer_message

    @task(name="generate_final_answer")  # type: ignore
    async def _generate_final_answer_stream(
        self, memory: LLMAgentMemory, ctx: RunContext[CtxT], call_id: str
    ) -> AsyncIterator[Event[Any]]:
        user_message = UserMessage.from_text(
            "Exceeded the maximum number of turns: provide a final answer now!",
        )
        memory.update([user_message])
        yield UserMessageEvent(
            proc_name=self.agent_name, call_id=call_id, data=user_message
        )
        if ctx.printer:
            ctx.printer.print_messages(
                [user_message], agent_name=self.agent_name, call_id=call_id
            )

        tool_choice = NamedToolChoice(name=self._final_answer_tool.name)
        async for event in self.generate_message_stream(
            memory, tool_choice=tool_choice, ctx=ctx, call_id=call_id
        ):
            yield event

        final_answer_message = self._extract_final_answer_from_tool_calls(memory)
        if final_answer_message is None:
            raise AgentFinalAnswerError(proc_name=self.agent_name, call_id=call_id)
        yield GenMessageEvent(
            proc_name=self.agent_name, call_id=call_id, data=final_answer_message
        )

    async def execute(
        self, memory: LLMAgentMemory, ctx: RunContext[CtxT], call_id: str
    ) -> AssistantMessage | Sequence[AssistantMessage]:
        # 1. Generate the first message:
        #    In ReAct mode, we generate the first message without tool calls
        #    to force the agent to plan its actions in a separate message.

        tool_choice: ToolChoice | None = None
        if self.tools:
            tool_choice = "none" if self._react_mode else "auto"
        gen_message = await self.generate_message(
            memory, tool_choice=tool_choice, ctx=ctx, call_id=call_id
        )
        if not self.tools:
            return gen_message

        turns = 0

        while True:
            # 2. Check if we should exit the tool call loop

            # If a final answer is not provided via a tool call, we use
            # _terminate_tool_call_loop to determine whether to exit the loop.
            if not self._final_answer_as_tool_call and self._terminate_tool_call_loop(
                memory.message_history, ctx=ctx, call_id=call_id, num_turns=turns
            ):
                return gen_message

            # If a final answer is provided via a tool call, we check
            # if the last message contains the corresponding tool call.
            # If it does, we exit the loop.
            if self._final_answer_as_tool_call:
                final_answer = self._extract_final_answer_from_tool_calls(memory)
                if final_answer is not None:
                    return final_answer

            # Exit if the maximum number of turns is reached
            if turns >= self.max_turns:
                # If a final answer is provided via a tool call, we force the
                # agent to use the final answer tool.
                # Otherwise, we simply return the last generated message.
                if self._final_answer_as_tool_call:
                    final_answer = await self._generate_final_answer(
                        memory, ctx=ctx, call_id=call_id
                    )
                else:
                    final_answer = gen_message
                logger.info(
                    f"Max turns reached: {self.max_turns}. Exiting the tool call loop."
                )
                return final_answer

            # 3. Call tools.

            if gen_message.tool_calls:
                await self.call_tools(
                    gen_message.tool_calls, memory=memory, ctx=ctx, call_id=call_id
                )

            # Apply memory management (e.g. compacting or pruning memory)
            self._manage_memory(memory, ctx=ctx, call_id=call_id, num_turns=turns)

            # 4. Generate the next message based on the updated memory.
            #    In ReAct mode, we set tool_choice to "none" if we just called tools,
            #    so the next message will be an observation/planning message with
            #    no immediate tool calls.

            if self._react_mode and gen_message.tool_calls:
                tool_choice = "none"
            elif gen_message.tool_calls:
                tool_choice = "auto"
            else:
                tool_choice = "required"

            gen_message = await self.generate_message(
                memory, tool_choice=tool_choice, ctx=ctx, call_id=call_id
            )

            turns += 1

    async def execute_stream(
        self,
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> AsyncIterator[Event[Any]]:
        tool_choice: ToolChoice = "none" if self._react_mode else "auto"
        gen_message: AssistantMessage | None = None
        async for event in self.generate_message_stream(
            memory, tool_choice=tool_choice, ctx=ctx, call_id=call_id
        ):
            if isinstance(event, GenMessageEvent):
                gen_message = event.data
            yield event
        if gen_message is None:
            return

        if not self.tools:
            return

        turns = 0

        while True:
            if not self._final_answer_as_tool_call and self._terminate_tool_call_loop(
                memory.message_history, ctx=ctx, call_id=call_id, num_turns=turns
            ):
                return

            if self._final_answer_as_tool_call:
                final_answer_message = self._extract_final_answer_from_tool_calls(
                    memory
                )
                if final_answer_message is not None:
                    yield GenMessageEvent(
                        proc_name=self.agent_name,
                        call_id=call_id,
                        data=final_answer_message,
                    )
                    return

            if turns >= self.max_turns:
                if self._final_answer_as_tool_call:
                    async for event in self._generate_final_answer_stream(
                        memory, ctx=ctx, call_id=call_id
                    ):
                        yield event
                logger.info(
                    f"Max turns reached: {self.max_turns}. Exiting the tool call loop."
                )
                return

            if gen_message.tool_calls:
                for tool_call in gen_message.tool_calls:
                    yield ToolCallEvent(
                        proc_name=self.agent_name, call_id=call_id, data=tool_call
                    )

                async for event in self.call_tools_stream(
                    gen_message.tool_calls, memory=memory, ctx=ctx, call_id=call_id
                ):
                    yield event

            self._manage_memory(memory, ctx=ctx, call_id=call_id, num_turns=turns)

            if self._react_mode and gen_message.tool_calls:
                tool_choice = "none"
            elif gen_message.tool_calls:
                tool_choice = "auto"
            else:
                tool_choice = "required"

            async for event in self.generate_message_stream(
                memory, tool_choice=tool_choice, ctx=ctx, call_id=call_id
            ):
                yield event
                if isinstance(event, GenMessageEvent):
                    gen_message = event.data

            turns += 1

    def get_final_answer_tool(self) -> BaseTool[BaseModel, None, Any]:
        class FinalAnswerTool(BaseTool[self._final_answer_type, None, Any]):
            name: str = "final_answer"
            description: str = (
                "You must call this tool to provide the final answer. "
                "DO NOT output your answer before calling the tool. "
            )

            async def run(
                self,
                inp: BaseModel,
                *,
                ctx: RunContext[Any] | None = None,
                call_id: str | None = None,
            ) -> None:
                return None

        return FinalAnswerTool()

    def _process_completion(
        self, completion: Completion, *, ctx: RunContext[CtxT], call_id: str
    ) -> None:
        ctx.completions[self.agent_name].append(completion)
        ctx.usage_tracker.update(
            agent_name=self.agent_name,
            completions=[completion],
            model_name=self.llm.model_name,
        )
        if ctx.printer:
            usages = [None] * (len(completion.messages) - 1) + [completion.usage]
            ctx.printer.print_messages(
                completion.messages,
                usages=usages,
                agent_name=self.agent_name,
                call_id=call_id,
            )



from collections.abc import Iterable
from typing import Any

from pydantic import BaseModel

from ..typing.completion import Completion, Usage
from ..typing.completion_chunk import CompletionChunk
from ..typing.content import Content
from ..typing.converters import Converters
from ..typing.message import AssistantMessage, SystemMessage, ToolMessage, UserMessage
from ..typing.tool import BaseTool, ToolChoice
from . import (
    OpenAIAssistantMessageParam,
    OpenAICompletion,
    OpenAICompletionChunk,
    OpenAICompletionMessage,
    OpenAIContentPartParam,
    OpenAISystemMessageParam,
    OpenAIToolChoiceOptionParam,
    OpenAIToolMessageParam,
    OpenAIToolParam,
    OpenAIUsage,
    OpenAIUserMessageParam,
)
from .completion_chunk_converters import from_api_completion_chunk
from .completion_converters import (
    from_api_completion,
    from_api_completion_usage,
    to_api_completion,
)
from .content_converters import from_api_content, to_api_content
from .message_converters import (
    from_api_assistant_message,
    from_api_system_message,
    from_api_tool_message,
    from_api_user_message,
    to_api_assistant_message,
    to_api_system_message,
    to_api_tool_message,
    to_api_user_message,
)
from .tool_converters import to_api_tool, to_api_tool_choice


class OpenAIConverters(Converters):
    @staticmethod
    def to_system_message(
        system_message: SystemMessage, **kwargs: Any
    ) -> OpenAISystemMessageParam:
        return to_api_system_message(system_message, **kwargs)

    @staticmethod
    def from_system_message(
        raw_message: OpenAISystemMessageParam, name: str | None = None, **kwargs: Any
    ) -> SystemMessage:
        return from_api_system_message(raw_message, name=name, **kwargs)

    @staticmethod
    def to_user_message(
        user_message: UserMessage, **kwargs: Any
    ) -> OpenAIUserMessageParam:
        return to_api_user_message(user_message, **kwargs)

    @staticmethod
    def from_user_message(
        raw_message: OpenAIUserMessageParam, name: str | None = None, **kwargs: Any
    ) -> UserMessage:
        return from_api_user_message(raw_message, name=name, **kwargs)

    @staticmethod
    def to_assistant_message(
        assistant_message: AssistantMessage, **kwargs: Any
    ) -> OpenAIAssistantMessageParam:
        return to_api_assistant_message(assistant_message, **kwargs)

    @staticmethod
    def from_completion_usage(raw_usage: OpenAIUsage, **kwargs: Any) -> Usage:
        return from_api_completion_usage(raw_usage, **kwargs)

    @staticmethod
    def from_assistant_message(
        raw_message: OpenAICompletionMessage, name: str | None = None, **kwargs: Any
    ) -> AssistantMessage:
        return from_api_assistant_message(raw_message, name=name, **kwargs)

    @staticmethod
    def to_tool_message(
        tool_message: ToolMessage, **kwargs: Any
    ) -> OpenAIToolMessageParam:
        return to_api_tool_message(tool_message, **kwargs)

    @staticmethod
    def from_tool_message(
        raw_message: OpenAIToolMessageParam, name: str | None = None, **kwargs: Any
    ) -> ToolMessage:
        return from_api_tool_message(raw_message, name=name, **kwargs)

    @staticmethod
    def to_tool(
        tool: BaseTool[BaseModel, Any, Any], strict: bool | None = None, **kwargs: Any
    ) -> OpenAIToolParam:
        return to_api_tool(tool, strict=strict, **kwargs)

    @staticmethod
    def to_tool_choice(
        tool_choice: ToolChoice, **kwargs: Any
    ) -> OpenAIToolChoiceOptionParam:
        return to_api_tool_choice(tool_choice, **kwargs)

    @staticmethod
    def to_content(content: Content, **kwargs: Any) -> Iterable[OpenAIContentPartParam]:
        return to_api_content(content, **kwargs)

    @staticmethod
    def from_content(
        raw_content: str | Iterable[OpenAIContentPartParam], **kwargs: Any
    ) -> Content:
        return from_api_content(raw_content, **kwargs)

    @staticmethod
    def to_completion(completion: Completion, **kwargs: Any) -> OpenAICompletion:
        return to_api_completion(completion, **kwargs)

    @staticmethod
    def from_completion(
        raw_completion: OpenAICompletion, name: str | None = None, **kwargs: Any
    ) -> Completion:
        return from_api_completion(raw_completion, name=name, **kwargs)

    @staticmethod
    def to_completion_chunk(
        chunk: CompletionChunk, **kwargs: Any
    ) -> OpenAICompletionChunk:
        raise NotImplementedError

    @staticmethod
    def from_completion_chunk(
        raw_chunk: OpenAICompletionChunk, name: str | None = None, **kwargs: Any
    ) -> CompletionChunk:
        return from_api_completion_chunk(raw_chunk, name=name, **kwargs)



from ..openai.completion_converters import from_api_completion_usage
from ..typing.completion_chunk import (
    CompletionChunk,
    CompletionChunkChoice,
    CompletionChunkChoiceDelta,
    CompletionChunkDeltaToolCall,
)
from . import LiteLLMChunkChoice, LiteLLMCompletionChunk


def from_api_completion_chunk(
    api_completion_chunk: LiteLLMCompletionChunk, name: str | None = None
) -> CompletionChunk:
    choices: list[CompletionChunkChoice] = []

    for api_choice in api_completion_chunk.choices:
        assert isinstance(api_choice, LiteLLMChunkChoice)

        api_delta = api_choice.delta

        delta = CompletionChunkChoiceDelta(
            tool_calls=[
                CompletionChunkDeltaToolCall(
                    id=tool_call.id,
                    index=tool_call.index,
                    tool_name=tool_call.function.name,
                    tool_arguments=tool_call.function.arguments,
                )
                for tool_call in (api_delta.tool_calls or [])
                if tool_call.function
            ],
            content=api_delta.content,  # type: ignore[assignment, arg-type]
            role=api_delta.role,  # type: ignore[assignment, arg-type]
            thinking_blocks=getattr(api_delta, "thinking_blocks", None),
            annotations=getattr(api_delta, "annotations", None),
            reasoning_content=getattr(api_delta, "reasoning_content", None),
            provider_specific_fields=api_delta.provider_specific_fields,
            refusal=getattr(api_delta, "refusal", None),
        )

        choice = CompletionChunkChoice(
            delta=delta,
            index=api_choice.index,
            finish_reason=api_choice.finish_reason,  # type: ignore[assignment, arg-type]
            logprobs=getattr(api_choice, "logprobs", None),
        )

        choices.append(choice)

    api_usage = getattr(api_completion_chunk, "usage", None)
    usage = None
    if api_usage is not None:
        usage = from_api_completion_usage(api_usage)
        hidden_params = getattr(api_completion_chunk, "_hidden_params", {})
        usage.cost = getattr(hidden_params, "response_cost", None)

    return CompletionChunk(
        id=api_completion_chunk.id,
        model=api_completion_chunk.model,
        name=name,
        created=api_completion_chunk.created,
        system_fingerprint=api_completion_chunk.system_fingerprint,
        choices=choices,
        usage=usage,
        provider_specific_fields=api_completion_chunk.provider_specific_fields,
        hidden_params=api_completion_chunk._hidden_params,  # type: ignore[union-attr]
        response_ms=getattr(api_completion_chunk, "_response_ms", None),
    )



import httpx
from pydantic import BaseModel, NonNegativeFloat, PositiveInt


class AsyncHTTPClientParams(BaseModel):
    timeout: NonNegativeFloat = 10
    max_connections: PositiveInt = 2000
    max_keepalive_connections: PositiveInt = 500
    keepalive_expiry: float | None = 5
    # proxy: str | None = None
    # follow_redirects: bool = False
    # trust_env: bool = True
    # auth: AuthTypes | None = (None,)
    # params: QueryParamTypes | None = (None,)
    # headers: HeaderTypes | None = (None,)
    # cookies: CookieTypes | None = (None,)


def create_simple_async_httpx_client(
    client_params: AsyncHTTPClientParams,
) -> httpx.AsyncClient:
    extra_params = client_params.model_dump(
        exclude={
            "timeout",
            "max_connections",
            "max_keepalive_connections",
            "keepalive_expiry",
        }
    )
    return httpx.AsyncClient(
        timeout=httpx.Timeout(client_params.timeout),
        limits=httpx.Limits(
            max_connections=client_params.max_connections,
            max_keepalive_connections=client_params.max_keepalive_connections,
            keepalive_expiry=client_params.keepalive_expiry,
        ),
        **extra_params,
    )



# from openai import APIResponseValidationError


class ProcRunError(Exception):
    def __init__(
        self, proc_name: str, call_id: str, message: str | None = None
    ) -> None:
        super().__init__(
            message
            or f"Processor run failed [proc_name: {proc_name}; call_id: {call_id}]."
        )
        self.proc_name = proc_name
        self.call_id = call_id


class ProcInputValidationError(ProcRunError):
    pass


class ProcOutputValidationError(ProcRunError):
    def __init__(
        self, schema: object, proc_name: str, call_id: str, message: str | None = None
    ):
        super().__init__(
            proc_name=proc_name,
            call_id=call_id,
            message=message
            or (
                "Processor output validation failed "
                f"[proc_name: {proc_name}; call_id: {call_id}]. "
                f"Expected type:\n{schema}"
            ),
        )


class AgentFinalAnswerError(ProcRunError):
    def __init__(
        self, proc_name: str, call_id: str, message: str | None = None
    ) -> None:
        super().__init__(
            proc_name=proc_name,
            call_id=call_id,
            message=message
            or "Final answer tool call did not return a final answer message "
            f"[proc_name={proc_name}; call_id={call_id}]",
        )
        self.message = message


class WorkflowConstructionError(Exception):
    pass


class PacketRoutingError(ProcRunError):
    def __init__(
        self,
        proc_name: str,
        call_id: str,
        selected_recipient: str | None = None,
        allowed_recipients: list[str] | None = None,
        message: str | None = None,
    ) -> None:
        default_message = (
            f"Selected recipient '{selected_recipient}' is not in the allowed "
            f"recipients: {allowed_recipients} "
            f"[proc_name={proc_name}; call_id={call_id}]"
        )
        super().__init__(
            proc_name=proc_name, call_id=call_id, message=message or default_message
        )
        self.selected_recipient = selected_recipient
        self.allowed_recipients = allowed_recipients


class RunnerError(Exception):
    pass


class PromptBuilderError(Exception):
    def __init__(self, proc_name: str, message: str | None = None) -> None:
        super().__init__(message or f"Prompt builder failed [proc_name={proc_name}]")
        self.proc_name = proc_name
        self.message = message


class SystemPromptBuilderError(PromptBuilderError):
    def __init__(self, proc_name: str, message: str | None = None) -> None:
        super().__init__(
            proc_name=proc_name,
            message=message
            or "System prompt builder failed to make system prompt "
            f"[proc_name={proc_name}]",
        )
        self.message = message


class InputPromptBuilderError(PromptBuilderError):
    def __init__(self, proc_name: str, message: str | None = None) -> None:
        super().__init__(
            proc_name=proc_name,
            message=message
            or "Input prompt builder failed to make input content "
            f"[proc_name={proc_name}]",
        )
        self.message = message


class PyJSONStringParsingError(Exception):
    def __init__(self, s: str, message: str | None = None) -> None:
        super().__init__(
            message
            or "Both ast.literal_eval and json.loads failed to parse the following "
            f"JSON/Python string:\n{s}"
        )
        self.s = s


class JSONSchemaValidationError(Exception):
    def __init__(self, s: str, schema: object, message: str | None = None) -> None:
        super().__init__(
            message
            or f"JSON schema validation failed for:\n{s}\nExpected type: {schema}"
        )
        self.s = s
        self.schema = schema


class CompletionError(Exception):
    pass


class CombineCompletionChunksError(Exception):
    pass


class LLMToolCallValidationError(Exception):
    def __init__(
        self, tool_name: str, tool_args: str, message: str | None = None
    ) -> None:
        super().__init__(
            message
            or f"Failed to validate tool call '{tool_name}' with arguments:"
            f"\n{tool_args}."
        )
        self.tool_name = tool_name
        self.tool_args = tool_args


class LLMResponseValidationError(JSONSchemaValidationError):
    def __init__(self, s: str, schema: object, message: str | None = None) -> None:
        super().__init__(
            s,
            schema,
            message
            or f"Failed to validate LLM response:\n{s}\nExpected type: {schema}",
        )



from collections.abc import Callable, Coroutine
from typing import Any, Concatenate, ParamSpec, TypeAlias, TypeVar

T = TypeVar("T")
R = TypeVar("R")
P = ParamSpec("P")

AsyncCallable: TypeAlias = Callable[P, Coroutine[Any, Any, R]]
AsyncFunction: TypeAlias = Callable[Concatenate[T, P], Coroutine[Any, Any, R]]
AsyncMethod: TypeAlias = Callable[Concatenate[Any, T, P], Coroutine[Any, Any, R]]
AsyncFunctionOrMethod: TypeAlias = AsyncFunction[T, P, R] | AsyncMethod[T, P, R]



import logging
from collections.abc import AsyncIterator, Sequence
from itertools import pairwise
from typing import Any, Generic, cast, final

from grasp_agents.tracing_decorators import workflow

from ..errors import WorkflowConstructionError
from ..packet_pool import Packet
from ..processors.base_processor import BaseProcessor
from ..run_context import CtxT, RunContext
from ..typing.events import Event, ProcPacketOutputEvent, WorkflowResultEvent
from ..typing.io import InT, OutT, ProcName
from .workflow_processor import WorkflowProcessor

logger = logging.getLogger(__name__)


class SequentialWorkflow(WorkflowProcessor[InT, OutT, CtxT], Generic[InT, OutT, CtxT]):
    def __init__(
        self,
        name: ProcName,
        subprocs: Sequence[BaseProcessor[Any, Any, Any, CtxT]],
        recipients: list[ProcName] | None = None,
        max_retries: int = 0,
    ) -> None:
        super().__init__(
            subprocs=subprocs,
            start_proc=subprocs[0],
            end_proc=subprocs[-1],
            name=name,
            recipients=recipients,
            max_retries=max_retries,
        )

        for prev_proc, proc in pairwise(subprocs):
            if prev_proc.out_type != proc.in_type:
                raise WorkflowConstructionError(
                    f"Output type {prev_proc.out_type} of subprocessor {prev_proc.name}"
                    f" does not match input type {proc.in_type} of subprocessor"
                    f" {proc.name}"
                )

    @workflow(name="workflow_run")  # type: ignore
    @final
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | Sequence[InT] | None = None,
        forgetful: bool = False,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT]:
        call_id = self._generate_call_id(call_id)

        packet = in_packet
        for subproc in self.subprocs:
            logger.info(f"\n[Running subprocessor {subproc.name}]\n")

            packet = await subproc.run(
                chat_inputs=chat_inputs,
                in_packet=packet,
                in_args=in_args,
                forgetful=forgetful,
                call_id=f"{call_id}/{subproc.name}",
                ctx=ctx,
            )
            chat_inputs = None
            in_args = None

            logger.info(f"\n[Finished running subprocessor {subproc.name}]\n")

        return cast("Packet[OutT]", packet)

    @workflow(name="workflow_run")  # type: ignore
    @final
    async def run_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | Sequence[InT] | None = None,
        forgetful: bool = False,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncIterator[Event[Any]]:
        call_id = self._generate_call_id(call_id)

        packet = in_packet
        for subproc in self.subprocs:
            logger.info(f"\n[Running subprocessor {subproc.name}]\n")

            async for event in subproc.run_stream(
                chat_inputs=chat_inputs,
                in_packet=packet,
                in_args=in_args,
                forgetful=forgetful,
                call_id=f"{call_id}/{subproc.name}",
                ctx=ctx,
            ):
                if isinstance(event, ProcPacketOutputEvent):
                    packet = event.data
                yield event

            chat_inputs = None
            in_args = None

            logger.info(f"\n[Finished running subprocessor {subproc.name}]\n")

        yield WorkflowResultEvent(
            data=cast("Packet[OutT]", packet), proc_name=self.name, call_id=call_id
        )



import logging
from abc import abstractmethod
from collections.abc import AsyncIterator, Mapping
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Required, cast

import httpx
from pydantic import BaseModel
from typing_extensions import TypedDict

# from .errors import LLMResponseValidationError, LLMToolCallValidationError
from .http_client import AsyncHTTPClientParams, create_simple_async_httpx_client
from .llm import LLM, LLMSettings
from .rate_limiting.rate_limiter import RateLimiter, limit_rate
from .typing.completion import Completion
from .typing.completion_chunk import CompletionChoice, CompletionChunk
from .typing.events import (
    CompletionChunkEvent,
    CompletionEvent,
    LLMStreamingErrorData,
    LLMStreamingErrorEvent,
)
from .typing.message import AssistantMessage, Messages
from .typing.tool import BaseTool, ToolChoice

logger = logging.getLogger(__name__)


class APIProvider(TypedDict, total=False):
    name: Required[str]
    base_url: str | None
    api_key: str | None
    # Wildcard patterns for model names that support response schema validation:
    response_schema_support: tuple[str, ...] | None


def make_refusal_completion(model_name: str, err: BaseException) -> Completion:
    failed_message = AssistantMessage(content=None, refusal=str(err))

    return Completion(
        model=model_name,
        choices=[CompletionChoice(message=failed_message, finish_reason=None, index=0)],
    )


class CloudLLMSettings(LLMSettings, total=False):
    extra_headers: dict[str, Any] | None
    extra_body: object | None
    extra_query: dict[str, Any] | None


LLMRateLimiter = RateLimiter[
    AssistantMessage
    | AsyncIterator[
        CompletionChunkEvent[CompletionChunk] | CompletionEvent | LLMStreamingErrorEvent
    ],
]


@dataclass(frozen=True)
class CloudLLM(LLM):
    llm_settings: CloudLLMSettings | None = None
    api_provider: APIProvider | None = None
    rate_limiter: LLMRateLimiter | None = None
    client_timeout: float = 60.0
    max_client_retries: int = 2  # HTTP client retries for network errors
    max_response_retries: int = (
        0  # LLM response retries: try to regenerate to pass validation
    )
    apply_response_schema_via_provider: bool = False
    apply_tool_call_schema_via_provider: bool = False
    async_http_client: httpx.AsyncClient | None = None
    async_http_client_params: AsyncHTTPClientParams | None = None

    def __post_init__(self) -> None:
        if self.rate_limiter is not None:
            logger.info(
                f"[{self.__class__.__name__}] Set rate limit to "
                f"{self.rate_limiter.rpm} RPM"
            )

        if self.apply_response_schema_via_provider:
            object.__setattr__(self, "apply_tool_call_schema_via_provider", True)

        if self.async_http_client is None and self.async_http_client_params is not None:
            logger.info(
                f"[{self.__class__.__name__}] Creating custom async HTTP client "
                f"with params:\n{self.async_http_client_params}"
            )
            object.__setattr__(
                self,
                "async_http_client",
                create_simple_async_httpx_client(self.async_http_client_params),
            )

    def _make_completion_kwargs(
        self,
        conversation: Messages,
        response_schema: Any | None = None,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
        tool_choice: ToolChoice | None = None,
        n_choices: int | None = None,
    ) -> dict[str, Any]:
        api_messages = [self.converters.to_message(m) for m in conversation]

        api_tools = None
        api_tool_choice = None
        if tools:
            strict = True if self.apply_tool_call_schema_via_provider else None
            api_tools = [
                self.converters.to_tool(t, strict=strict) for t in tools.values()
            ]
            if tool_choice is not None:
                api_tool_choice = self.converters.to_tool_choice(tool_choice)

        api_llm_settings = deepcopy(self.llm_settings or {})

        return dict(
            api_messages=api_messages,
            api_tools=api_tools,
            api_tool_choice=api_tool_choice,
            api_response_schema=response_schema,
            n_choices=n_choices,
            **api_llm_settings,
        )

    @abstractmethod
    async def _get_completion(
        self,
        api_messages: list[Any],
        *,
        api_tools: list[Any] | None = None,
        api_tool_choice: Any | None = None,
        api_response_schema: type | None = None,
        n_choices: int | None = None,
        **api_llm_settings: Any,
    ) -> Any:
        pass

    @abstractmethod
    async def _get_completion_stream(
        self,
        api_messages: list[Any],
        *,
        api_tools: list[Any] | None = None,
        api_tool_choice: Any | None = None,
        api_response_schema: type | None = None,
        n_choices: int | None = None,
        **api_llm_settings: Any,
    ) -> AsyncIterator[Any]:
        pass

    @limit_rate
    async def _generate_completion_once(
        self,
        conversation: Messages,
        *,
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
        tool_choice: ToolChoice | None = None,
        n_choices: int | None = None,
    ) -> Completion:
        completion_kwargs = self._make_completion_kwargs(
            conversation=conversation,
            response_schema=response_schema,
            tools=tools,
            tool_choice=tool_choice,
            n_choices=n_choices,
        )

        if not self.apply_response_schema_via_provider:
            completion_kwargs.pop("api_response_schema", None)
        api_completion = await self._get_completion(**completion_kwargs)

        completion = self.converters.from_completion(api_completion, name=self.model_id)

        # if not self.apply_response_schema_via_provider:
        self._validate_response(
            completion,
            response_schema=response_schema,
            response_schema_by_xml_tag=response_schema_by_xml_tag,
        )
        # if not self.apply_tool_call_schema_via_provider and tools is not None:
        if tools is not None:
            self._validate_tool_calls(completion, tools=tools)

        return completion

    async def generate_completion(
        self,
        conversation: Messages,
        *,
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
        tool_choice: ToolChoice | None = None,
        n_choices: int | None = None,
        proc_name: str | None = None,
        call_id: str | None = None,
    ) -> Completion:
        n_attempt = 0
        while n_attempt <= self.max_response_retries:
            try:
                return await self._generate_completion_once(
                    conversation,  # type: ignore[return]
                    response_schema=response_schema,
                    response_schema_by_xml_tag=response_schema_by_xml_tag,
                    tools=tools,
                    tool_choice=tool_choice,
                    n_choices=n_choices,
                )
            # except (LLMResponseValidationError, LLMToolCallValidationError) as err:
            except Exception as err:
                n_attempt += 1

                if n_attempt > self.max_response_retries:
                    if n_attempt == 1:
                        logger.warning(f"\nCloudLLM completion failed:\n{err}")
                    if n_attempt > 1:
                        logger.warning(
                            f"\nCloudLLM completion failed after retrying:\n{err}"
                        )
                    raise err
                    # return make_refusal_completion(self._model_name, err)

                logger.warning(
                    f"\nCloudLLM completion failed (retry attempt {n_attempt}):\n{err}"
                )

        return make_refusal_completion(
            self.model_name,
            Exception("Unexpected error: retry loop exited without returning"),
        )

    @limit_rate
    async def _generate_completion_stream_once(
        self,
        conversation: Messages,
        *,
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
        tool_choice: ToolChoice | None = None,
        n_choices: int | None = None,
        proc_name: str | None = None,
        call_id: str | None = None,
    ) -> AsyncIterator[CompletionChunkEvent[CompletionChunk] | CompletionEvent]:
        completion_kwargs = self._make_completion_kwargs(
            conversation=conversation,
            response_schema=response_schema,
            tools=tools,
            tool_choice=tool_choice,
            n_choices=n_choices,
        )
        if not self.apply_response_schema_via_provider:
            completion_kwargs.pop("api_response_schema", None)

        api_stream = self._get_completion_stream(**completion_kwargs)
        api_stream = cast("AsyncIterator[Any]", api_stream)

        async def iterator() -> AsyncIterator[
            CompletionChunkEvent[CompletionChunk] | CompletionEvent
        ]:
            api_completion_chunks: list[Any] = []

            async for api_completion_chunk in api_stream:
                api_completion_chunks.append(api_completion_chunk)
                completion_chunk = self.converters.from_completion_chunk(
                    api_completion_chunk, name=self.model_id
                )

                yield CompletionChunkEvent(
                    data=completion_chunk, proc_name=proc_name, call_id=call_id
                )

            api_completion = self.combine_completion_chunks(
                api_completion_chunks, response_schema=response_schema, tools=tools
            )
            completion = self.converters.from_completion(
                api_completion, name=self.model_id
            )

            yield CompletionEvent(data=completion, proc_name=proc_name, call_id=call_id)

            # if not self.apply_response_schema_via_provider:
            self._validate_response(
                completion,
                response_schema=response_schema,
                response_schema_by_xml_tag=response_schema_by_xml_tag,
            )
            # if not self.apply_tool_call_schema_via_provider and tools is not None:
            if tools is not None:
                self._validate_tool_calls(completion, tools=tools)

        return iterator()

    async def generate_completion_stream(  # type: ignore[override]
        self,
        conversation: Messages,
        *,
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
        tool_choice: ToolChoice | None = None,
        n_choices: int | None = None,
        proc_name: str | None = None,
        call_id: str | None = None,
    ) -> AsyncIterator[
        CompletionChunkEvent[CompletionChunk] | CompletionEvent | LLMStreamingErrorEvent
    ]:
        n_attempt = 0
        while n_attempt <= self.max_response_retries:
            try:
                async for event in await self._generate_completion_stream_once(  # type: ignore[return]
                    conversation,  # type: ignore[arg-type]
                    response_schema=response_schema,
                    response_schema_by_xml_tag=response_schema_by_xml_tag,
                    tools=tools,
                    tool_choice=tool_choice,
                    n_choices=n_choices,
                    proc_name=proc_name,
                    call_id=call_id,
                ):
                    yield event
                return
            # except (LLMResponseValidationError, LLMToolCallValidationError) as err:
            except Exception as err:
                err_data = LLMStreamingErrorData(
                    error=err, model_name=self.model_name, model_id=self.model_id
                )
                yield LLMStreamingErrorEvent(
                    data=err_data, proc_name=proc_name, call_id=call_id
                )

                n_attempt += 1
                if n_attempt > self.max_response_retries:
                    if n_attempt == 1:
                        logger.warning(f"\nCloudLLM completion failed:\n{err}")
                    if n_attempt > 1:
                        logger.warning(
                            f"\nCloudLLM completion failed after retrying:\n{err}"
                        )
                    raise err
                    # refusal_completion = make_refusal_completion(
                    #     self.model_name, err
                    # )
                    # yield CompletionEvent(
                    #     data=refusal_completion,
                    #     proc_name=proc_name,
                    #     call_id=call_id,
                    # )

                logger.warning(
                    f"\nCloudLLM completion failed (retry attempt {n_attempt}):\n{err}"
                )



import logging
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Callable, Coroutine, Sequence
from functools import wraps
from typing import (
    Any,
    ClassVar,
    Generic,
    Protocol,
    TypeVar,
    cast,
    final,
)
from uuid import uuid4

from pydantic import BaseModel, TypeAdapter
from pydantic import ValidationError as PydanticValidationError

from ..errors import (
    PacketRoutingError,
    ProcInputValidationError,
    ProcOutputValidationError,
    ProcRunError,
)
from ..generics_utils import AutoInstanceAttributesMixin
from ..memory import DummyMemory, MemT
from ..packet import Packet
from ..run_context import CtxT, RunContext
from ..typing.events import (
    DummyEvent,
    Event,
    ProcStreamingErrorData,
    ProcStreamingErrorEvent,
)
from ..typing.io import InT, OutT, ProcName
from ..typing.tool import BaseTool

logger = logging.getLogger(__name__)


F = TypeVar("F", bound=Callable[..., Coroutine[Any, Any, Packet[Any]]])
F_stream = TypeVar("F_stream", bound=Callable[..., AsyncIterator[Event[Any]]])


def with_retry(func: F) -> F:
    @wraps(func)
    async def wrapper(
        self: "BaseProcessor[Any, Any, Any, Any]", *args: Any, **kwargs: Any
    ) -> Packet[Any]:
        none_packet = Packet(payloads=[None], sender=self.name)
        call_id = kwargs.get("call_id", "unknown")
        for n_attempt in range(self.max_retries + 1):
            try:
                return await func(self, *args, **kwargs)
            except Exception as err:
                err_message = (
                    f"\nProcessor run failed [proc_name={self.name}; call_id={call_id}]"
                )
                if n_attempt == self.max_retries:
                    if self.max_retries == 0:
                        logger.warning(f"{err_message}:\n{err}")
                    else:
                        logger.warning(f"{err_message} after retrying:\n{err}")
                    # raise ProcRunError(proc_name=self.name, call_id=call_id) from err
                    return none_packet

                logger.warning(f"{err_message} (retry attempt {n_attempt + 1}):\n{err}")
        # This part should not be reachable due to the raise in the loop
        # raise ProcRunError(proc_name=self.name, call_id=call_id)
        return none_packet

    return cast("F", wrapper)


def with_retry_stream(func: F_stream) -> F_stream:
    @wraps(func)
    async def wrapper(
        self: "BaseProcessor[Any, Any, Any, Any]", *args: Any, **kwargs: Any
    ) -> AsyncIterator[Event[Any]]:
        call_id = kwargs.get("call_id", "unknown")
        for n_attempt in range(self.max_retries + 1):
            try:
                async for event in func(self, *args, **kwargs):
                    yield event
                return
            except Exception as err:
                err_data = ProcStreamingErrorData(error=err, call_id=call_id)
                yield ProcStreamingErrorEvent(
                    data=err_data, proc_name=self.name, call_id=call_id
                )
                err_message = (
                    "\nStreaming processor run failed "
                    f"[proc_name={self.name}; call_id={call_id}]"
                )
                if n_attempt == self.max_retries:
                    if self.max_retries == 0:
                        logger.warning(f"{err_message}:\n{err}")
                    else:
                        logger.warning(f"{err_message} after retrying:\n{err}")
                    raise ProcRunError(proc_name=self.name, call_id=call_id) from err

                logger.warning(f"{err_message} (retry attempt {n_attempt}):\n{err}")

    return cast("F_stream", wrapper)


_OutT_contra = TypeVar("_OutT_contra", contravariant=True)


class RecipientSelector(Protocol[_OutT_contra, CtxT]):
    def __call__(
        self, output: _OutT_contra, *, ctx: RunContext[CtxT]
    ) -> Sequence[ProcName] | None: ...


class BaseProcessor(AutoInstanceAttributesMixin, ABC, Generic[InT, OutT, MemT, CtxT]):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        name: ProcName,
        max_retries: int = 0,
        recipients: Sequence[ProcName] | None = None,
        **kwargs: Any,
    ) -> None:
        self._in_type: type[InT]
        self._out_type: type[OutT]

        super().__init__()

        self._name: ProcName = name
        self._memory: MemT = cast("MemT", DummyMemory())
        self._max_retries: int = max_retries

        self.recipients = recipients

        self.recipient_selector: RecipientSelector[OutT, CtxT] | None
        if not hasattr(type(self), "recipient_selector"):
            self.recipient_selector = None

    @property
    def in_type(self) -> type[InT]:
        return self._in_type

    @property
    def out_type(self) -> type[OutT]:
        return self._out_type

    @property
    def name(self) -> ProcName:
        return self._name

    @property
    def memory(self) -> MemT:
        return self._memory

    @property
    def max_retries(self) -> int:
        return self._max_retries

    def _generate_call_id(self, call_id: str | None) -> str:
        _new_id = str(uuid4())[:6] + "_" + self.name
        return _new_id if call_id is None else call_id + "/" + _new_id

    def _validate_inputs(
        self,
        call_id: str,
        chat_inputs: Any | None = None,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
    ) -> list[InT] | None:
        mult_inputs_err_message = (
            "Only one of chat_inputs, in_args, or in_message must be provided."
        )
        err_kwargs = {"proc_name": self.name, "call_id": call_id}

        if chat_inputs is not None and in_args is not None:
            raise ProcInputValidationError(
                message=mult_inputs_err_message, **err_kwargs
            )
        if chat_inputs is not None and in_packet is not None:
            raise ProcInputValidationError(
                message=mult_inputs_err_message, **err_kwargs
            )
        if in_args is not None and in_packet is not None:
            raise ProcInputValidationError(
                message=mult_inputs_err_message, **err_kwargs
            )

        if in_packet is not None and not in_packet.payloads:
            raise ProcInputValidationError(
                message="in_packet must contain at least one payload.", **err_kwargs
            )
        if in_args is not None and not in_args:
            raise ProcInputValidationError(
                message="in_args must contain at least one argument.", **err_kwargs
            )

        if chat_inputs is not None:
            return None

        resolved_args: list[InT]

        if isinstance(in_args, list):
            _in_args = cast("list[Any]", in_args)
            if all(isinstance(x, self.in_type) for x in _in_args):
                resolved_args = cast("list[InT]", _in_args)
            elif isinstance(_in_args, self.in_type):
                resolved_args = cast("list[InT]", [_in_args])
            else:
                raise ProcInputValidationError(
                    message=f"in_args are neither of type {self.in_type} "
                    f"nor a sequence of {self.in_type}.",
                    **err_kwargs,
                )

        elif in_args is not None:
            resolved_args = cast("list[InT]", [in_args])

        else:
            assert in_packet is not None
            resolved_args = cast("list[InT]", in_packet.payloads)

        try:
            for args in resolved_args:
                TypeAdapter(self._in_type).validate_python(args)
        except PydanticValidationError as err:
            raise ProcInputValidationError(message=str(err), **err_kwargs) from err

        return resolved_args

    def _validate_output(self, out_payload: OutT, call_id: str) -> OutT:
        if out_payload is None:
            return out_payload
        try:
            return TypeAdapter(self._out_type).validate_python(out_payload)
        except PydanticValidationError as err:
            raise ProcOutputValidationError(
                schema=self._out_type, proc_name=self.name, call_id=call_id
            ) from err

    def _validate_recipients(
        self, recipients: Sequence[ProcName] | None, call_id: str
    ) -> None:
        for r in recipients or []:
            if r not in (self.recipients or []):
                raise PacketRoutingError(
                    proc_name=self.name,
                    call_id=call_id,
                    selected_recipient=r,
                    allowed_recipients=cast("list[str]", self.recipients),
                )

    @final
    def _select_recipients(
        self, output: OutT, ctx: RunContext[CtxT]
    ) -> Sequence[ProcName] | None:
        if self.recipient_selector:
            return self.recipient_selector(output=output, ctx=ctx)

        return self.recipients

    def add_recipient_selector(
        self, func: RecipientSelector[OutT, CtxT]
    ) -> RecipientSelector[OutT, CtxT]:
        self.recipient_selector = func

        return func

    @abstractmethod
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
        forgetful: bool = False,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT]:
        pass

    @abstractmethod
    async def run_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
        forgetful: bool = False,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncIterator[Event[Any]]:
        yield DummyEvent()

    @final
    def as_tool(
        self, tool_name: str, tool_description: str
    ) -> BaseTool[InT, OutT, Any]:  # type: ignore[override]
        # TODO: stream tools
        processor_instance = self
        in_type = processor_instance.in_type
        out_type = processor_instance.out_type
        if not issubclass(in_type, BaseModel):
            raise TypeError(
                "Cannot create a tool from an agent with "
                f"non-BaseModel input type: {in_type}"
            )

        class ProcessorTool(BaseTool[in_type, out_type, Any]):
            name: str = tool_name
            description: str = tool_description

            async def run(
                self,
                inp: InT,
                *,
                call_id: str | None = None,
                ctx: RunContext[CtxT] | None = None,
            ) -> OutT:
                result = await processor_instance.run(
                    in_args=inp, forgetful=True, call_id=call_id, ctx=ctx
                )

                return result.payloads[0]

        return ProcessorTool()



import asyncio
import functools
import logging
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from time import monotonic
from typing import Generic, overload

from .types import AsyncCallable, P, R
from .utils import split_pos_args

logger = logging.getLogger(__name__)


MAX_RPM = 1e10

RateLimDecorator = Callable[[AsyncCallable[P, R]], AsyncCallable[P, R]]


@dataclass
class RateLimiterState:
    next_request_time: float = 0.0


class RateLimiter(Generic[R]):
    def __init__(self, rpm: float, max_concurrency: int = 200):
        self._rpm = rpm
        self._max_concurrency = max_concurrency

        self._lock = asyncio.Lock()
        self._state = RateLimiterState(next_request_time=0.0)
        self._semaphore = asyncio.Semaphore(self._max_concurrency)

    async def process(self, func_partial: AsyncCallable[..., R]) -> R:
        async with self._semaphore:
            async with self._lock:
                now = monotonic()
                if now < self._state.next_request_time:
                    await asyncio.sleep(self._state.next_request_time - now)
                self._state.next_request_time = monotonic() + 1.01 * 60.0 / self._rpm
            return await func_partial()

    @property
    def rpm(self) -> float:
        return self._rpm

    @rpm.setter
    def rpm(self, value: float) -> None:
        self._rpm = value

    @property
    def max_concurrency(self) -> int:
        return self._max_concurrency

    @property
    def state(self) -> RateLimiterState:
        return self._state


@overload
def limit_rate(
    call: AsyncCallable[P, R], rate_limiter: RateLimiter[R] | None = None
) -> AsyncCallable[P, R]: ...


@overload
def limit_rate(
    call: None = None, rate_limiter: RateLimiter[R] | None = None
) -> RateLimDecorator[P, R]: ...


def limit_rate(
    call: AsyncCallable[P, R] | None = None, rate_limiter: RateLimiter[R] | None = None
) -> AsyncCallable[P, R] | RateLimDecorator[P, R]:
    if call is None:
        return functools.partial(limit_rate, rate_limiter=rate_limiter)

    @functools.wraps(call)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        self_obj, _, _ = split_pos_args(call, args)  # type: ignore
        call_partial = partial(call, *args, **kwargs)
        _rate_limiter = rate_limiter or getattr(self_obj, "rate_limiter", None)
        if _rate_limiter is None:
            return await call_partial()

        return await _rate_limiter.process(func_partial=call_partial)

    return wrapper



# pyright: reportUnusedImport=false

from litellm.types.utils import ChatCompletionMessageToolCall as LiteLLMToolCall
from litellm.types.utils import Choices as LiteLLMChoice
from litellm.types.utils import Function as LiteLLMFunction
from litellm.types.utils import Message as LiteLLMCompletionMessage
from litellm.types.utils import ModelResponse as LiteLLMCompletion
from litellm.types.utils import ModelResponseStream as LiteLLMCompletionChunk
from litellm.types.utils import StreamingChoices as LiteLLMChunkChoice
from litellm.types.utils import Usage as LiteLLMUsage
from openai._streaming import (
    AsyncStream as OpenAIAsyncStream,  # type: ignore[import] # noqa: PLC2701
)
from openai.types import CompletionUsage as OpenAIUsage
from openai.types.chat.chat_completion import ChatCompletion as OpenAICompletion
from openai.types.chat.chat_completion import (
    ChoiceLogprobs as OpenAIChoiceLogprobs,
)
from openai.types.chat.chat_completion_assistant_message_param import (
    ChatCompletionAssistantMessageParam as OpenAIAssistantMessageParam,
)
from openai.types.chat.chat_completion_chunk import (
    ChatCompletionChunk as OpenAICompletionChunk,
)
from openai.types.chat.chat_completion_chunk import (
    Choice as OpenAIChunkChoice,
)
from openai.types.chat.chat_completion_chunk import (
    ChoiceDelta as OpenAIChunkChoiceDelta,
)
from openai.types.chat.chat_completion_chunk import (
    ChoiceDeltaToolCall as OpenAIChunkChoiceDeltaToolCall,
)
from openai.types.chat.chat_completion_content_part_image_param import (
    ChatCompletionContentPartImageParam as OpenAIContentPartImageParam,
)
from openai.types.chat.chat_completion_content_part_image_param import (
    ImageURL as OpenAIImageURL,
)
from openai.types.chat.chat_completion_content_part_param import (
    ChatCompletionContentPartParam as OpenAIContentPartParam,
)
from openai.types.chat.chat_completion_content_part_text_param import (
    ChatCompletionContentPartTextParam as OpenAIContentPartTextParam,
)
from openai.types.chat.chat_completion_developer_message_param import (
    ChatCompletionDeveloperMessageParam as OpenAIDeveloperMessageParam,
)
from openai.types.chat.chat_completion_function_message_param import (
    ChatCompletionFunctionMessageParam as OpenAIFunctionMessageParam,
)
from openai.types.chat.chat_completion_message import (
    ChatCompletionMessage as OpenAICompletionMessage,
)
from openai.types.chat.chat_completion_message_param import (
    ChatCompletionMessageParam as OpenAIMessageParam,
)
from openai.types.chat.chat_completion_message_tool_call_param import (
    ChatCompletionMessageToolCallParam as OpenAIToolCallParam,
)
from openai.types.chat.chat_completion_message_tool_call_param import (
    Function as OpenAIToolCallFunction,
)
from openai.types.chat.chat_completion_named_tool_choice_param import (
    ChatCompletionNamedToolChoiceParam as OpenAINamedToolChoiceParam,
)
from openai.types.chat.chat_completion_named_tool_choice_param import (
    Function as OpenAINamedToolChoiceFunction,
)
from openai.types.chat.chat_completion_prediction_content_param import (
    ChatCompletionPredictionContentParam as OpenAIPredictionContentParam,
)
from openai.types.chat.chat_completion_stream_options_param import (
    ChatCompletionStreamOptionsParam as OpenAIStreamOptionsParam,
)
from openai.types.chat.chat_completion_system_message_param import (
    ChatCompletionSystemMessageParam as OpenAISystemMessageParam,
)
from openai.types.chat.chat_completion_tool_choice_option_param import (
    ChatCompletionToolChoiceOptionParam as OpenAIToolChoiceOptionParam,
)
from openai.types.chat.chat_completion_tool_message_param import (
    ChatCompletionToolMessageParam as OpenAIToolMessageParam,
)
from openai.types.chat.chat_completion_tool_param import (
    ChatCompletionToolParam as OpenAIToolParam,
)
from openai.types.chat.chat_completion_user_message_param import (
    ChatCompletionUserMessageParam as OpenAIUserMessageParam,
)
from openai.types.chat.parsed_chat_completion import (
    ParsedChatCompletion as OpenAIParsedCompletion,
)
from openai.types.chat.parsed_chat_completion import (
    ParsedChatCompletionMessage as OpenAIParsedCompletionMessage,
)
from openai.types.chat.parsed_chat_completion import (
    ParsedChoice as OpenAIParsedChoice,
)
from openai.types.shared_params.function_definition import (
    FunctionDefinition as OpenAIFunctionDefinition,
)

from .lite_llm import LiteLLM, LiteLLMSettings

__all__ = ["LiteLLM", "LiteLLMSettings"]



from ..errors import CompletionError
from ..typing.completion import Completion, CompletionChoice, Usage
from . import OpenAICompletion, OpenAIUsage
from .message_converters import from_api_assistant_message


def from_api_completion_usage(api_usage: OpenAIUsage) -> Usage:
    reasoning_tokens = None
    cached_tokens = None

    if api_usage.completion_tokens_details is not None:
        reasoning_tokens = api_usage.completion_tokens_details.reasoning_tokens
    if api_usage.prompt_tokens_details is not None:
        cached_tokens = api_usage.prompt_tokens_details.cached_tokens

    input_tokens = api_usage.prompt_tokens - (cached_tokens or 0)
    output_tokens = api_usage.completion_tokens - (reasoning_tokens or 0)

    return Usage(
        input_tokens=input_tokens,
        output_tokens=output_tokens,
        reasoning_tokens=reasoning_tokens,
        cached_tokens=cached_tokens,
    )


def from_api_completion(
    api_completion: OpenAICompletion, name: str | None = None
) -> Completion:
    choices: list[CompletionChoice] = []

    if api_completion.choices is None:  # type: ignore
        # Some providers return None for the choices when there is an error
        # TODO: add custom error types
        raise CompletionError(
            f"Completion API error: {getattr(api_completion, 'error', None)}"
        )
    for api_choice in api_completion.choices:
        finish_reason = api_choice.finish_reason

        # Some providers return None for the message when finish_reason is other than "stop"
        if api_choice.message is None:  # type: ignore
            raise CompletionError(
                f"API returned None for message with finish_reason: {finish_reason}"
            )

        message = from_api_assistant_message(api_choice.message, name=name)

        choices.append(
            CompletionChoice(
                index=api_choice.index,
                message=message,
                finish_reason=finish_reason,
                logprobs=api_choice.logprobs,
            )
        )

    usage = (
        from_api_completion_usage(api_completion.usage)
        if api_completion.usage
        else None
    )

    return Completion(
        id=api_completion.id,
        created=api_completion.created,
        system_fingerprint=api_completion.system_fingerprint,
        usage=usage,
        choices=choices,
        model=api_completion.model,
        name=name,
    )


def to_api_completion(completion: Completion) -> OpenAICompletion:
    raise NotImplementedError



from collections.abc import Sequence
from typing import Annotated, Any, Generic, Literal, TypeVar
from uuid import uuid4

from pydantic import AfterValidator, BaseModel, ConfigDict, Field

from .typing.io import ProcName

START_PROC_NAME: Literal["*START*"] = "*START*"

_PayloadT_co = TypeVar("_PayloadT_co", covariant=True)


class Packet(BaseModel, Generic[_PayloadT_co]):
    id: str = Field(default_factory=lambda: str(uuid4())[:8])
    payloads: Sequence[_PayloadT_co]
    sender: ProcName
    recipients: Sequence[ProcName] | None = None

    model_config = ConfigDict(extra="forbid")

    def __repr__(self) -> str:
        _to = ", ".join(self.recipients) if self.recipients else "None"
        return (
            f"{self.__class__.__name__}:\n"
            f"ID: {self.id}\n"
            f"From: {self.sender}\n"
            f"To: {_to}\n"
            f"Payloads: {len(self.payloads)}"
        )


def _check_recipients_length(v: Sequence[ProcName] | None) -> Sequence[ProcName] | None:
    if v is not None and len(v) != 1:
        raise ValueError("recipients must contain exactly one item")
    return v


class StartPacket(Packet[_PayloadT_co]):
    chat_inputs: Any | None = "start"
    sender: ProcName = Field(default=START_PROC_NAME, frozen=True)
    payloads: Sequence[_PayloadT_co] = Field(default=(), frozen=True)
    recipients: Annotated[
        Sequence[ProcName] | None,
        AfterValidator(_check_recipients_length),
    ] = None



from collections.abc import AsyncIterator, Mapping, Sequence
from pathlib import Path
from typing import Any, ClassVar, Generic, Protocol, TypeVar, cast, final

from pydantic import BaseModel

from .llm import LLM
from .llm_agent_memory import LLMAgentMemory, MemoryPreparator
from .llm_policy_executor import (
    LLMPolicyExecutor,
    MemoryManager,
    ToolCallLoopTerminator,
)
from .processors.parallel_processor import ParallelProcessor
from .prompt_builder import (
    InputContentBuilder,
    PromptBuilder,
    SystemPromptBuilder,
)
from .run_context import CtxT, RunContext
from .typing.content import Content, ImageData
from .typing.events import (
    Event,
    ProcPayloadOutputEvent,
    SystemMessageEvent,
    UserMessageEvent,
)
from .typing.io import InT, LLMPrompt, OutT, ProcName
from .typing.message import Message, Messages, SystemMessage, UserMessage
from .typing.tool import BaseTool
from .utils import get_prompt, validate_obj_from_json_or_py_string

_InT_contra = TypeVar("_InT_contra", contravariant=True)
_OutT_co = TypeVar("_OutT_co", covariant=True)


class OutputParser(Protocol[_InT_contra, _OutT_co, CtxT]):
    def __call__(
        self,
        conversation: Messages,
        *,
        in_args: _InT_contra | None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> _OutT_co: ...


class LLMAgent(
    ParallelProcessor[InT, OutT, LLMAgentMemory, CtxT],
    Generic[InT, OutT, CtxT],
):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        name: ProcName,
        *,
        # LLM
        llm: LLM,
        # Tools
        tools: list[BaseTool[Any, Any, CtxT]] | None = None,
        # Input prompt template (combines user and received arguments)
        in_prompt: LLMPrompt | None = None,
        in_prompt_path: str | Path | None = None,
        # System prompt template
        sys_prompt: LLMPrompt | None = None,
        sys_prompt_path: str | Path | None = None,
        # LLM response validation
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        # Agent loop settings
        max_turns: int = 100,
        react_mode: bool = False,
        final_answer_as_tool_call: bool = False,
        # Agent memory management
        reset_memory_on_run: bool = False,
        # Agent run retries
        max_retries: int = 0,
        # Multi-agent routing
        recipients: Sequence[ProcName] | None = None,
    ) -> None:
        super().__init__(name=name, recipients=recipients, max_retries=max_retries)

        # Agent memory

        self._memory: LLMAgentMemory = LLMAgentMemory()
        self._reset_memory_on_run = reset_memory_on_run

        self.memory_preparator: MemoryPreparator | None
        if not hasattr(type(self), "memory_preparator"):
            self.memory_preparator = None

        self.output_parser: OutputParser[InT, OutT, CtxT] | None
        if not hasattr(type(self), "output_parser"):
            self.output_parser = None

        # LLM policy executor

        if issubclass(self._out_type, BaseModel):
            final_answer_type = self._out_type
        elif not final_answer_as_tool_call:
            final_answer_type = BaseModel
        else:
            raise TypeError(
                "Final answer type must be a subclass of BaseModel if "
                "final_answer_as_tool_call is True."
            )

        self._used_default_llm_response_schema: bool = False
        if (
            response_schema is None
            and tools is None
            and not hasattr(type(self), "output_parser")
        ):
            response_schema = self.out_type
            self._used_default_llm_response_schema = True

        self._policy_executor: LLMPolicyExecutor[CtxT] = LLMPolicyExecutor[CtxT](
            agent_name=self.name,
            llm=llm,
            tools=tools,
            response_schema=response_schema,
            response_schema_by_xml_tag=response_schema_by_xml_tag,
            max_turns=max_turns,
            react_mode=react_mode,
            final_answer_type=final_answer_type,
            final_answer_as_tool_call=final_answer_as_tool_call,
        )

        # Prompt builder

        sys_prompt = get_prompt(prompt_text=sys_prompt, prompt_path=sys_prompt_path)
        in_prompt = get_prompt(prompt_text=in_prompt, prompt_path=in_prompt_path)

        self._prompt_builder = PromptBuilder[self.in_type, CtxT](
            agent_name=self._name, sys_prompt=sys_prompt, in_prompt=in_prompt
        )

        self._register_overridden_handlers()

    @property
    def llm(self) -> LLM:
        return self._policy_executor.llm

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, Any, CtxT]]:
        return self._policy_executor.tools

    @property
    def max_turns(self) -> int:
        return self._policy_executor.max_turns

    @property
    def sys_prompt(self) -> LLMPrompt | None:
        return self._prompt_builder.sys_prompt

    @property
    def in_prompt(self) -> LLMPrompt | None:
        return self._prompt_builder.in_prompt

    @final
    def _prepare_memory(
        self,
        memory: LLMAgentMemory,
        *,
        in_args: InT | None = None,
        sys_prompt: LLMPrompt | None = None,
        ctx: RunContext[Any],
        call_id: str,
    ) -> None:
        if self.memory_preparator:
            return self.memory_preparator(
                memory=memory,
                in_args=in_args,
                sys_prompt=sys_prompt,
                ctx=ctx,
                call_id=call_id,
            )

    def _memorize_inputs(
        self,
        memory: LLMAgentMemory,
        *,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        in_args: InT | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> tuple[SystemMessage | None, UserMessage | None]:
        formatted_sys_prompt = self._prompt_builder.build_system_prompt(
            ctx=ctx, call_id=call_id
        )

        system_message: SystemMessage | None = None
        if self._reset_memory_on_run or memory.is_empty:
            memory.reset(formatted_sys_prompt)
            if formatted_sys_prompt is not None:
                system_message = cast("SystemMessage", memory.message_history[0])
        else:
            self._prepare_memory(
                memory=memory,
                in_args=in_args,
                sys_prompt=formatted_sys_prompt,
                ctx=ctx,
                call_id=call_id,
            )

        input_message = self._prompt_builder.build_input_message(
            chat_inputs=chat_inputs, in_args=in_args, ctx=ctx, call_id=call_id
        )
        if input_message:
            memory.update([input_message])

        return system_message, input_message

    def parse_output_default(self, conversation: Messages) -> OutT:
        return validate_obj_from_json_or_py_string(
            str(conversation[-1].content or ""),
            schema=self._out_type,
            from_substring=False,
            strip_language_markdown=True,
        )

    def _parse_output(
        self,
        conversation: Messages,
        *,
        in_args: InT | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> OutT:
        if self.output_parser:
            return self.output_parser(
                conversation=conversation, in_args=in_args, ctx=ctx, call_id=call_id
            )

        return self.parse_output_default(conversation)

    async def _process(
        self,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        *,
        in_args: InT | None = None,
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> OutT:
        system_message, input_message = self._memorize_inputs(
            memory=memory,
            chat_inputs=chat_inputs,
            in_args=in_args,
            ctx=ctx,
            call_id=call_id,
        )
        if system_message:
            self._print_messages([system_message], ctx=ctx, call_id=call_id)
        if input_message:
            self._print_messages([input_message], ctx=ctx, call_id=call_id)

        await self._policy_executor.execute(memory, ctx=ctx, call_id=call_id)

        return self._parse_output(
            conversation=memory.message_history,
            in_args=in_args,
            ctx=ctx,
            call_id=call_id,
        )

    async def _process_stream(
        self,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        *,
        in_args: InT | None = None,
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> AsyncIterator[Event[Any]]:
        system_message, input_message = self._memorize_inputs(
            memory=memory,
            chat_inputs=chat_inputs,
            in_args=in_args,
            ctx=ctx,
            call_id=call_id,
        )
        if system_message:
            self._print_messages([system_message], ctx=ctx, call_id=call_id)
            yield SystemMessageEvent(
                data=system_message, proc_name=self.name, call_id=call_id
            )
        if input_message:
            self._print_messages([input_message], ctx=ctx, call_id=call_id)
            yield UserMessageEvent(
                data=input_message, proc_name=self.name, call_id=call_id
            )

        async for event in self._policy_executor.execute_stream(
            memory, ctx=ctx, call_id=call_id
        ):
            yield event

        output = self._parse_output(
            conversation=memory.message_history,
            in_args=in_args,
            ctx=ctx,
            call_id=call_id,
        )
        yield ProcPayloadOutputEvent(data=output, proc_name=self.name, call_id=call_id)

    def _print_messages(
        self, messages: Sequence[Message], ctx: RunContext[CtxT], call_id: str
    ) -> None:
        if ctx.printer:
            ctx.printer.print_messages(messages, agent_name=self.name, call_id=call_id)

    # -- Override these methods in subclasses if needed --

    def _register_overridden_handlers(self) -> None:
        cur_cls = type(self)
        base_cls = LLMAgent[Any, Any, Any]

        # Prompt builder

        if cur_cls.system_prompt_builder is not base_cls.system_prompt_builder:
            self._prompt_builder.system_prompt_builder = self.system_prompt_builder

        if cur_cls.input_content_builder is not base_cls.input_content_builder:
            self._prompt_builder.input_content_builder = self.input_content_builder

        # Policy executor

        if cur_cls.tool_call_loop_terminator is not base_cls.tool_call_loop_terminator:
            self._policy_executor.tool_call_loop_terminator = (
                self.tool_call_loop_terminator
            )

        if cur_cls.memory_manager is not base_cls.memory_manager:
            self._policy_executor.memory_manager = self.memory_manager

    def system_prompt_builder(self, ctx: RunContext[CtxT], call_id: str) -> str | None:
        if self._prompt_builder.system_prompt_builder is not None:
            return self._prompt_builder.system_prompt_builder(ctx=ctx, call_id=call_id)
        raise NotImplementedError("System prompt builder is not implemented.")

    def input_content_builder(
        self, in_args: InT, ctx: RunContext[CtxT], call_id: str
    ) -> Content:
        if self._prompt_builder.input_content_builder is not None:
            return self._prompt_builder.input_content_builder(
                in_args=in_args, ctx=ctx, call_id=call_id
            )
        raise NotImplementedError("Input content builder is not implemented.")

    def tool_call_loop_terminator(
        self,
        conversation: Messages,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> bool:
        if self._policy_executor.tool_call_loop_terminator is not None:
            return self._policy_executor.tool_call_loop_terminator(
                conversation=conversation, ctx=ctx, call_id=call_id, **kwargs
            )
        raise NotImplementedError("Tool call loop terminator is not implemented.")

    def memory_manager(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> None:
        if self._policy_executor.memory_manager is not None:
            return self._policy_executor.memory_manager(
                memory=memory, ctx=ctx, call_id=call_id, **kwargs
            )
        raise NotImplementedError("Memory manager is not implemented.")

    # Decorators for custom implementations as an alternative to overriding methods

    def add_system_prompt_builder(
        self, func: SystemPromptBuilder[CtxT]
    ) -> SystemPromptBuilder[CtxT]:
        self._prompt_builder.system_prompt_builder = func

        return func

    def add_input_content_builder(
        self, func: InputContentBuilder[InT, CtxT]
    ) -> InputContentBuilder[InT, CtxT]:
        self._prompt_builder.input_content_builder = func

        return func

    def add_memory_manager(self, func: MemoryManager[CtxT]) -> MemoryManager[CtxT]:
        self._policy_executor.memory_manager = func

        return func

    def add_tool_call_loop_terminator(
        self, func: ToolCallLoopTerminator[CtxT]
    ) -> ToolCallLoopTerminator[CtxT]:
        self._policy_executor.tool_call_loop_terminator = func

        return func

    def add_output_parser(
        self, func: OutputParser[InT, OutT, CtxT]
    ) -> OutputParser[InT, OutT, CtxT]:
        if self._used_default_llm_response_schema:
            self._policy_executor.response_schema = None
        self.output_parser = func

        return func

    def add_memory_preparator(self, func: MemoryPreparator) -> MemoryPreparator:
        self.memory_preparator = func
        return func



import logging
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Mapping
from dataclasses import dataclass, field
from typing import Any, ClassVar
from uuid import uuid4

from pydantic import BaseModel
from typing_extensions import TypedDict

from grasp_agents.typing.completion_chunk import CompletionChunk
from grasp_agents.utils import (
    validate_obj_from_json_or_py_string,
    validate_tagged_objs_from_json_or_py_string,
)

from .errors import (
    JSONSchemaValidationError,
    LLMResponseValidationError,
    LLMToolCallValidationError,
)
from .typing.completion import Completion
from .typing.converters import Converters
from .typing.events import (
    AnnotationsChunkEvent,
    AnnotationsEndEvent,
    AnnotationsStartEvent,
    CompletionChunkEvent,
    # CompletionEndEvent,
    CompletionEvent,
    CompletionStartEvent,
    LLMStateChangeEvent,
    LLMStreamingErrorEvent,
    # RefusalChunkEvent,
    ResponseChunkEvent,
    ResponseEndEvent,
    ResponseStartEvent,
    ThinkingChunkEvent,
    ThinkingEndEvent,
    ThinkingStartEvent,
    ToolCallChunkEvent,
    ToolCallEndEvent,
    ToolCallStartEvent,
)
from .typing.message import Messages
from .typing.tool import BaseTool, ToolChoice

logger = logging.getLogger(__name__)


LLMStreamGenerator = AsyncIterator[
    CompletionChunkEvent[CompletionChunk]
    | CompletionEvent
    | LLMStateChangeEvent[Any]
    | LLMStreamingErrorEvent
]


class LLMSettings(TypedDict, total=False):
    max_completion_tokens: int | None
    temperature: float | None
    top_p: float | None
    seed: int | None


@dataclass(frozen=True)
class LLM(ABC):
    model_name: str
    converters: ClassVar[Converters]
    llm_settings: LLMSettings | None = None
    model_id: str = field(default_factory=lambda: str(uuid4())[:8])

    def _validate_response(
        self,
        completion: Completion,
        response_schema: Any | None,
        response_schema_by_xml_tag: Mapping[str, Any] | None,
    ) -> None:
        if response_schema and response_schema_by_xml_tag:
            raise ValueError(
                "Only one of response_schema and response_schema_by_xml_tag can be "
                "provided, but not both."
            )
        parsing_params = {
            "from_substring": False,
            "strip_language_markdown": True,
        }
        try:
            for message in completion.messages:
                if not message.tool_calls:
                    if response_schema:
                        validate_obj_from_json_or_py_string(
                            message.content or "",
                            schema=response_schema,
                            **parsing_params,
                        )
                    elif response_schema_by_xml_tag:
                        validate_tagged_objs_from_json_or_py_string(
                            message.content or "",
                            schema_by_xml_tag=response_schema_by_xml_tag,
                            **parsing_params,
                        )
        except JSONSchemaValidationError as exc:
            raise LLMResponseValidationError(
                exc.s, exc.schema, message=str(exc)
            ) from exc

    def _validate_tool_calls(
        self, completion: Completion, tools: Mapping[str, BaseTool[BaseModel, Any, Any]]
    ) -> None:
        parsing_params = {
            "from_substring": False,
            "strip_language_markdown": True,
        }
        for message in completion.messages:
            if message.tool_calls:
                for tool_call in message.tool_calls:
                    tool_name = tool_call.tool_name
                    tool_arguments = tool_call.tool_arguments

                    available_tool_names = list(tools) if tools else []
                    if tool_name not in available_tool_names or not tools:
                        raise LLMToolCallValidationError(
                            tool_name,
                            tool_arguments,
                            message=f"Tool '{tool_name}' is not available in the LLM "
                            f"tools (available: {available_tool_names})",
                        )
                    tool = tools[tool_name]
                    try:
                        validate_obj_from_json_or_py_string(
                            tool_arguments, schema=tool.in_type, **parsing_params
                        )
                    except JSONSchemaValidationError as exc:
                        raise LLMToolCallValidationError(
                            tool_name, tool_arguments
                        ) from exc

    @staticmethod
    async def postprocess_event_stream(
        stream: LLMStreamGenerator,
    ) -> LLMStreamGenerator:
        prev_completion_id: str | None = None
        chunk_op_evt: CompletionChunkEvent[CompletionChunk] | None = None
        response_op_evt: ResponseChunkEvent | None = None
        thinking_op_evt: ThinkingChunkEvent | None = None
        annotations_op_evt: AnnotationsChunkEvent | None = None
        tool_calls_op_evt: ToolCallChunkEvent | None = None

        def _close_open_events(
            _event: CompletionChunkEvent[CompletionChunk] | None = None,
        ) -> list[LLMStateChangeEvent[Any]]:
            nonlocal \
                chunk_op_evt, \
                thinking_op_evt, \
                tool_calls_op_evt, \
                response_op_evt, \
                annotations_op_evt

            events: list[LLMStateChangeEvent[Any]] = []

            if not isinstance(_event, ThinkingChunkEvent) and thinking_op_evt:
                events.append(ThinkingEndEvent.from_chunk_event(thinking_op_evt))
                thinking_op_evt = None

            if not isinstance(_event, ToolCallChunkEvent) and tool_calls_op_evt:
                events.append(ToolCallEndEvent.from_chunk_event(tool_calls_op_evt))
                tool_calls_op_evt = None

            if not isinstance(_event, ResponseChunkEvent) and response_op_evt:
                events.append(ResponseEndEvent.from_chunk_event(response_op_evt))
                response_op_evt = None

            if not isinstance(_event, AnnotationsChunkEvent) and annotations_op_evt:
                events.append(AnnotationsEndEvent.from_chunk_event(annotations_op_evt))
                annotations_op_evt = None

            return events

        async for event in stream:
            if isinstance(event, CompletionChunkEvent) and not isinstance(
                event, LLMStateChangeEvent
            ):
                chunk = event.data
                if len(chunk.choices) != 1:
                    raise ValueError(
                        "Expected exactly one choice in completion chunk, "
                        f"got {len(chunk.choices)}"
                    )

                new_completion = chunk.id != prev_completion_id

                if new_completion:
                    for close_event in _close_open_events():
                        yield close_event

                    chunk_op_evt = event
                    yield CompletionStartEvent.from_chunk_event(event)

                sub_events = event.split_into_specialized()

                for sub_event in sub_events:
                    for close_event in _close_open_events(sub_event):
                        yield close_event

                    if isinstance(sub_event, ThinkingChunkEvent):
                        if not thinking_op_evt:
                            thinking_op_evt = sub_event
                            yield ThinkingStartEvent.from_chunk_event(sub_event)
                        yield sub_event

                    if isinstance(sub_event, ToolCallChunkEvent):
                        tc = sub_event.data.tool_call
                        if tc.id:
                            # Tool call ID is not None only for the first chunk of a tool call
                            if tool_calls_op_evt:
                                yield ToolCallEndEvent.from_chunk_event(
                                    tool_calls_op_evt
                                )
                                tool_calls_op_evt = None
                            tool_calls_op_evt = sub_event
                            yield ToolCallStartEvent.from_chunk_event(sub_event)
                        yield sub_event

                    if isinstance(sub_event, ResponseChunkEvent):
                        if not response_op_evt:
                            response_op_evt = sub_event
                            yield ResponseStartEvent.from_chunk_event(sub_event)
                        yield sub_event

                    if isinstance(sub_event, AnnotationsChunkEvent):
                        if not annotations_op_evt:
                            annotations_op_evt = sub_event
                            yield AnnotationsStartEvent.from_chunk_event(sub_event)
                        yield sub_event

                prev_completion_id = chunk.id

            else:
                for close_event in _close_open_events():
                    yield close_event

                yield event

    @abstractmethod
    async def generate_completion(
        self,
        conversation: Messages,
        *,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        tool_choice: ToolChoice | None = None,
        n_choices: int | None = None,
        proc_name: str | None = None,
        call_id: str | None = None,
    ) -> Completion:
        pass

    @abstractmethod
    async def generate_completion_stream(
        self,
        conversation: Messages,
        *,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        tool_choice: ToolChoice | None = None,
        n_choices: int | None = None,
        proc_name: str | None = None,
        call_id: str | None = None,
    ) -> AsyncIterator[
        CompletionChunkEvent[CompletionChunk] | CompletionEvent | LLMStreamingErrorEvent
    ]:
        pass

    @abstractmethod
    def combine_completion_chunks(
        self,
        completion_chunks: list[Any],
        response_schema: Any | None = None,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
    ) -> Any:
        raise NotImplementedError



import time
from collections import defaultdict
from collections.abc import Sequence
from typing import Any
from uuid import uuid4

from litellm import ChatCompletionAnnotation as LiteLLMAnnotation
from litellm.types.utils import ChoiceLogprobs as LiteLLMChoiceLogprobs
from openai.types.chat.chat_completion import (
    ChoiceLogprobs as OpenAIChoiceLogprobs,
)
from openai.types.chat.chat_completion_chunk import (
    ChoiceLogprobs as OpenAIChunkChoiceLogprobs,
)
from openai.types.chat.chat_completion_token_logprob import (
    ChatCompletionTokenLogprob as OpenAITokenLogprob,
)
from pydantic import BaseModel, Field, ValidationError, field_validator

from ..errors import CombineCompletionChunksError
from .completion import Completion, CompletionChoice, FinishReason, Usage
from .message import (
    AssistantMessage,
    RedactedThinkingBlock,
    Role,
    ThinkingBlock,
    ToolCall,
)


class CompletionChunkDeltaToolCall(BaseModel):
    id: str | None
    index: int
    tool_name: str | None
    tool_arguments: str | None


class CompletionChunkChoiceDelta(BaseModel):
    content: str | None = None
    refusal: str | None = None
    role: Role | None = None
    tool_calls: list[CompletionChunkDeltaToolCall] | None = None
    reasoning_content: str | None = None
    thinking_blocks: list[ThinkingBlock | RedactedThinkingBlock] | None = None
    annotations: list[LiteLLMAnnotation] | None = None
    provider_specific_fields: dict[str, Any] | None = None

    @property
    def thinking_delta(self) -> "CompletionChunkChoiceDelta | None":
        return (
            CompletionChunkChoiceDelta(
                reasoning_content=self.reasoning_content,
                thinking_blocks=self.thinking_blocks,
                role=self.role,
                provider_specific_fields=self.provider_specific_fields,
            )
            if self.reasoning_content or self.thinking_blocks
            else None
        )

    @property
    def tool_call_deltas(self) -> "list[CompletionChunkChoiceDelta] | None":
        return (
            [
                CompletionChunkChoiceDelta(
                    tool_calls=[tool_call],
                    role=self.role,
                    provider_specific_fields=self.provider_specific_fields,
                )
                for tool_call in self.tool_calls
            ]
            if self.tool_calls
            else None
        )

    @property
    def response_delta(self) -> "CompletionChunkChoiceDelta | None":
        return (
            CompletionChunkChoiceDelta(
                content=self.content,
                role=self.role,
                provider_specific_fields=self.provider_specific_fields,
            )
            if self.content
            else None
        )

    @property
    def annotations_delta(self) -> "CompletionChunkChoiceDelta | None":
        return (
            CompletionChunkChoiceDelta(
                annotations=self.annotations,
                role=self.role,
                provider_specific_fields=self.provider_specific_fields,
            )
            if self.annotations
            else None
        )

    @property
    def refusal_delta(self) -> "CompletionChunkChoiceDelta | None":
        return (
            CompletionChunkChoiceDelta(
                refusal=self.refusal,
                role=self.role,
                provider_specific_fields=self.provider_specific_fields,
            )
            if self.refusal
            else None
        )


class CompletionChunkChoice(BaseModel):
    delta: CompletionChunkChoiceDelta
    finish_reason: FinishReason | None
    index: int
    logprobs: OpenAIChunkChoiceLogprobs | LiteLLMChoiceLogprobs | Any | None = None


class CompletionChunk(BaseModel):
    id: str = Field(default_factory=lambda: str(uuid4())[:8])
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str | None
    name: str | None = None
    system_fingerprint: str | None = None
    choices: list[CompletionChunkChoice]
    usage: Usage | None = None
    # LiteLLM-specific fields
    provider_specific_fields: dict[str, Any] | None = None
    response_ms: float | None = None
    hidden_params: dict[str, Any] | None = None

    def split_into_specialized(
        self,
    ) -> "list[CompletionChunk]":
        if len(self.choices) != 1:
            raise ValidationError(
                "CompletionChunk must have exactly one choice for specialization."
            )
        delta = self.choices[0].delta

        specialized_chunks: list[CompletionChunk] = []

        thinking_delta = delta.thinking_delta
        tool_call_deltas = delta.tool_call_deltas
        response_delta = delta.response_delta
        annotations_delta = delta.annotations_delta
        refusal_delta = delta.refusal_delta

        if thinking_delta is not None:
            new_choice = self.choices[0].model_copy(update={"delta": thinking_delta})
            new_chunk = self.model_copy(update={"choices": [new_choice]})
            specialized_chunks.append(
                ThinkingChunk.model_validate(new_chunk.model_dump())
            )

        if tool_call_deltas:
            for delta_tool_call in tool_call_deltas:
                new_choice = self.choices[0].model_copy(
                    update={"delta": delta_tool_call}
                )
                new_chunk = self.model_copy(update={"choices": [new_choice]})
                specialized_chunks.append(
                    ToolCallChunk.model_validate(new_chunk.model_dump())
                )

        if response_delta is not None:
            new_choice = self.choices[0].model_copy(update={"delta": response_delta})
            new_chunk = self.model_copy(update={"choices": [new_choice]})
            specialized_chunks.append(
                ResponseChunk.model_validate(new_chunk.model_dump())
            )

        if annotations_delta is not None:
            new_choice = self.choices[0].model_copy(update={"delta": annotations_delta})
            new_chunk = self.model_copy(update={"choices": [new_choice]})
            specialized_chunks.append(
                AnnotationsChunk.model_validate(new_chunk.model_dump())
            )

        if refusal_delta is not None:
            new_choice = self.choices[0].model_copy(update={"delta": refusal_delta})
            new_chunk = self.model_copy(update={"choices": [new_choice]})
            specialized_chunks.append(
                RefusalChunk.model_validate(new_chunk.model_dump())
            )

        return specialized_chunks


class ResponseChunk(CompletionChunk):
    @field_validator("choices")
    @classmethod
    def validate_response_chunk(
        cls, choices: list[CompletionChunkChoice]
    ) -> list[CompletionChunkChoice]:
        if len(choices) != 1:
            raise ValidationError("ResponseChunk must have exactly one choice.")

        delta = choices[0].delta

        if not delta.content:
            raise ValidationError("ResponseChunk must have content in deltas.")

        if (
            delta.reasoning_content is not None
            or delta.thinking_blocks is not None
            or delta.tool_calls is not None
            or delta.refusal is not None
            or delta.annotations is not None
        ):
            raise ValidationError(
                "ResponseChunk should not have reasoning content, thinking blocks, "
                "tool calls, refusal, or annotations in deltas."
            )

        return choices

    @property
    def response(self) -> str:
        assert self.choices[0].delta.content
        return self.choices[0].delta.content


class ThinkingChunk(CompletionChunk):
    @field_validator("choices")
    @classmethod
    def validate_thinking_chunk(
        cls, choices: list[CompletionChunkChoice]
    ) -> list[CompletionChunkChoice]:
        if len(choices) != 1:
            raise ValidationError("ThinkingChunk must have exactly one choice.")

        delta = choices[0].delta

        if not (delta.thinking_blocks or delta.reasoning_content):
            raise ValidationError(
                "ThinkingChunk must have reasoning content or "
                "at least one thinking block."
            )
        if (
            delta.content is not None
            or delta.tool_calls is not None
            or delta.refusal is not None
            or delta.annotations is not None
        ):
            raise ValidationError(
                "ThinkingChunk should not have content, tool calls, "
                "refusal, or annotations in deltas."
            )

        return choices

    @property
    def thinking(self) -> str | list[ThinkingBlock | RedactedThinkingBlock]:
        delta = self.choices[0].delta
        if delta.reasoning_content:
            return delta.reasoning_content
        if delta.thinking_blocks:
            return delta.thinking_blocks
        raise ValueError("ThinkingChunk has no reasoning_content or thinking_blocks")


class ToolCallChunk(CompletionChunk):
    @field_validator("choices")
    @classmethod
    def validate_tool_call_chunk(
        cls, choices: list[CompletionChunkChoice]
    ) -> list[CompletionChunkChoice]:
        if len(choices) != 1:
            raise ValidationError("ToolCallChunk must have exactly one choice.")

        delta = choices[0].delta

        if not delta.tool_calls:
            raise ValidationError("ToolCallChunk must have tool calls in deltas.")
        if len(delta.tool_calls) != 1:
            raise ValidationError(
                "ToolCallChunk must have exactly one tool call in deltas."
            )

        if (
            delta.reasoning_content is not None
            or delta.thinking_blocks is not None
            or delta.content is not None
            or delta.refusal is not None
            or delta.annotations is not None
        ):
            raise ValidationError(
                "ToolCallChunk should not have reasoning content, thinking blocks, "
                "content, refusal, or annotations in deltas."
            )

        return choices

    @property
    def tool_call(self) -> CompletionChunkDeltaToolCall:
        assert self.choices[0].delta.tool_calls is not None
        return self.choices[0].delta.tool_calls[0]


class AnnotationsChunk(CompletionChunk):
    @field_validator("choices")
    @classmethod
    def validate_annotations_chunk(
        cls, choices: list[CompletionChunkChoice]
    ) -> list[CompletionChunkChoice]:
        if len(choices) != 1:
            raise ValidationError("AnnotationsChunk must have exactly one choice.")

        delta = choices[0].delta

        if not delta.annotations:
            raise ValidationError("AnnotationsChunk must have annotations in deltas.")

        if (
            delta.reasoning_content is not None
            or delta.thinking_blocks is not None
            or delta.content is not None
            or delta.tool_calls is not None
            or delta.refusal is not None
        ):
            raise ValidationError(
                "AnnotationsChunk should not have reasoning content, thinking blocks, "
                "content, tool calls, or refusal in deltas."
            )

        return choices

    @property
    def annotations(self) -> list[LiteLLMAnnotation]:
        assert self.choices[0].delta.annotations is not None
        return self.choices[0].delta.annotations


class RefusalChunk(CompletionChunk):
    @field_validator("choices")
    @classmethod
    def validate_refusal_chunk(
        cls, choices: list[CompletionChunkChoice]
    ) -> list[CompletionChunkChoice]:
        if len(choices) != 1:
            raise ValidationError("RefusalChunk must have exactly one choice.")

        delta = choices[0].delta

        if not delta.refusal:
            raise ValidationError("RefusalChunk must have refusal in deltas.")

        if (
            delta.reasoning_content is not None
            or delta.thinking_blocks is not None
            or delta.content is not None
            or delta.tool_calls is not None
            or delta.annotations is not None
        ):
            raise ValidationError(
                "RefusalChunk should not have reasoning content, thinking blocks, "
                "content, tool calls, or annotations in deltas."
            )

        return choices

    @property
    def refusal(self) -> str | None:
        return self.choices[0].delta.refusal


def combine_completion_chunks(chunks: list[CompletionChunk]) -> Completion:
    if not chunks:
        raise CombineCompletionChunksError(
            "Cannot combine an empty list of completion chunks."
        )

    model_list = {chunk.model for chunk in chunks}
    if len(model_list) > 1:
        raise CombineCompletionChunksError("All chunks must have the same model.")
    model = model_list.pop()

    name_list = {chunk.name for chunk in chunks}
    if len(name_list) > 1:
        raise CombineCompletionChunksError("All chunks must have the same name.")
    name = name_list.pop()

    system_fingerprints_list = {chunk.system_fingerprint for chunk in chunks}
    if len(system_fingerprints_list) > 1:
        raise CombineCompletionChunksError(
            "All chunks must have the same system fingerprint."
        )
    system_fingerprint = system_fingerprints_list.pop()

    created_list = [chunk.created for chunk in chunks]
    created = max(created_list)

    # Usage is found in the last completion chunk if requested
    usage = chunks[-1].usage

    logp_contents_per_choice: defaultdict[int, list[OpenAITokenLogprob]] = defaultdict(
        list
    )
    logp_refusals_per_choice: defaultdict[int, list[OpenAITokenLogprob]] = defaultdict(
        list
    )
    logprobs_per_choice: defaultdict[int, OpenAIChoiceLogprobs | None] = defaultdict(
        lambda: None
    )
    thinking_blocks_per_choice: defaultdict[
        int, list[ThinkingBlock | RedactedThinkingBlock]
    ] = defaultdict(list)
    annotations_per_choice: defaultdict[int, list[LiteLLMAnnotation]] = defaultdict(
        list
    )

    finish_reasons_per_choice: defaultdict[int, FinishReason | None] = defaultdict(
        lambda: None
    )

    contents_per_choice: defaultdict[int, str] = defaultdict(lambda: "")
    reasoning_contents_per_choice: defaultdict[int, str] = defaultdict(lambda: "")
    refusals_per_choice: defaultdict[int, str] = defaultdict(lambda: "")

    tool_calls_per_choice: defaultdict[
        int, Sequence[CompletionChunkDeltaToolCall] | None
    ] = defaultdict(lambda: None)

    messages_per_choice: dict[int, AssistantMessage] = {}

    for chunk in chunks:
        for choice in chunk.choices:
            index = choice.index

            # Concatenate content and refusal tokens for each choice
            contents_per_choice[index] += choice.delta.content or ""
            reasoning_contents_per_choice[index] += choice.delta.reasoning_content or ""
            refusals_per_choice[index] += choice.delta.refusal or ""

            # Concatenate logprobs for content and refusal tokens for each choice
            if choice.logprobs is not None:
                logp_contents_per_choice[index].extend(choice.logprobs.content or [])  # type: ignore
                logp_refusals_per_choice[index].extend(choice.logprobs.refusal or [])  # type: ignore
                thinking_blocks_per_choice[index].extend(
                    choice.delta.thinking_blocks or []
                )
                annotations_per_choice[index].extend(choice.delta.annotations or [])

            # Take the last finish reason for each choice
            finish_reasons_per_choice[index] = choice.finish_reason

            # Tool calls should be in the last chunk for each choice
            tool_calls_per_choice[index] = choice.delta.tool_calls

    for index in finish_reasons_per_choice:
        tool_calls: list[ToolCall] = []
        if tool_calls_per_choice[index] is not None:
            for _tool_call in tool_calls_per_choice[index]:  # type: ignore
                if (
                    _tool_call.id is None
                    or _tool_call.tool_name is None
                    or _tool_call.tool_arguments is None
                ):
                    raise CombineCompletionChunksError(
                        "Completion chunk tool calls must have id, tool_name, "
                        "and tool_arguments set."
                    )
                tool_calls.append(
                    ToolCall(
                        id=_tool_call.id,
                        tool_name=_tool_call.tool_name,
                        tool_arguments=_tool_call.tool_arguments,
                    )
                )

        messages_per_choice[index] = AssistantMessage(
            name=name,
            content=contents_per_choice[index] or "<empty>",
            reasoning_content=(reasoning_contents_per_choice[index] or None),
            thinking_blocks=(thinking_blocks_per_choice[index] or None),
            annotations=(annotations_per_choice[index] or None),
            refusal=(refusals_per_choice[index] or None),
            tool_calls=(tool_calls or None),
        )

        if logp_contents_per_choice[index] or logp_refusals_per_choice[index]:
            logprobs_per_choice[index] = OpenAIChoiceLogprobs(
                content=logp_contents_per_choice[index],
                refusal=logp_refusals_per_choice[index],
            )

    choices = [
        CompletionChoice(
            index=index,
            message=message,
            finish_reason=finish_reasons_per_choice[index],
            logprobs=logprobs_per_choice[index],
        )
        for index, message in messages_per_choice.items()
    ]

    return Completion(
        model=model,
        name=name,
        created=created,
        system_fingerprint=system_fingerprint,
        choices=choices,
        usage=usage,
    )



from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Sequence
from typing import Any, Generic

from ..errors import WorkflowConstructionError
from ..packet import Packet
from ..processors.base_processor import BaseProcessor, RecipientSelector
from ..run_context import CtxT, RunContext
from ..typing.events import DummyEvent, Event
from ..typing.io import InT, OutT, ProcName


class WorkflowProcessor(
    BaseProcessor[InT, OutT, Any, CtxT],
    ABC,
    Generic[InT, OutT, CtxT],
):
    def __init__(
        self,
        name: ProcName,
        subprocs: Sequence[BaseProcessor[Any, Any, Any, CtxT]],
        start_proc: BaseProcessor[InT, Any, Any, CtxT],
        end_proc: BaseProcessor[Any, OutT, Any, CtxT],
        recipients: Sequence[ProcName] | None = None,
        max_retries: int = 0,
    ) -> None:
        super().__init__(name=name, recipients=recipients, max_retries=max_retries)

        self._in_type = start_proc.in_type
        self._out_type = end_proc.out_type

        if len(subprocs) < 2:
            raise WorkflowConstructionError("At least two subprocessors are required")
        if start_proc not in subprocs:
            raise WorkflowConstructionError(
                "Start subprocessor must be in the subprocessors list"
            )
        if end_proc not in subprocs:
            raise WorkflowConstructionError(
                "End subprocessor must be in the subprocessors list"
            )

        self._subprocs = subprocs
        self._start_proc = start_proc
        self._end_proc = end_proc

        self.recipients = recipients
        if hasattr(type(self), "recipient_selector"):
            self._end_proc.recipient_selector = self.recipient_selector

    def add_recipient_selector(
        self, func: RecipientSelector[OutT, CtxT]
    ) -> RecipientSelector[OutT, CtxT]:
        self._end_proc.recipient_selector = func
        self.recipient_selector = func

        return func

    @property
    def recipients(self) -> Sequence[ProcName] | None:
        return self._end_proc.recipients

    @recipients.setter
    def recipients(self, value: Sequence[ProcName] | None) -> None:
        if hasattr(self, "_end_proc"):
            self._end_proc.recipients = value

    @property
    def subprocs(self) -> Sequence[BaseProcessor[Any, Any, Any, CtxT]]:
        return self._subprocs

    @property
    def start_proc(self) -> BaseProcessor[InT, Any, Any, CtxT]:
        return self._start_proc

    @property
    def end_proc(self) -> BaseProcessor[Any, OutT, Any, CtxT]:
        return self._end_proc

    def _generate_subproc_call_id(
        self, call_id: str | None, subproc: BaseProcessor[Any, Any, Any, CtxT]
    ) -> str | None:
        return f"{self._generate_call_id(call_id)}/{subproc.name}"

    @abstractmethod
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | Sequence[InT] | None = None,
        ctx: RunContext[CtxT] | None = None,
        forgetful: bool = False,
        call_id: str | None = None,
    ) -> Packet[OutT]:
        pass

    @abstractmethod
    async def run_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | Sequence[InT] | None = None,
        ctx: RunContext[CtxT] | None = None,
        forgetful: bool = False,
        call_id: str | None = None,
    ) -> AsyncIterator[Event[Any]]:
        yield DummyEvent()
